# standard libraries
import copy
import logging
import math
import unittest

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.swift import Application
from nion.swift import DisplayPanel
from nion.swift import ImageCanvasItem
from nion.swift.model import DataItem
from nion.swift.model import Graphics
from nion.swift.test import TestContext
from nion.ui import CanvasItem
from nion.ui import TestUI
from nion.utils import Geometry


class TestGraphicsClass(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)

    def tearDown(self):
        pass

    def __get_mapping(self):
        return ImageCanvasItem.ImageCanvasItemMapping((1000, 1000), Geometry.FloatRect.from_tlbr(0, 0, 1000, 1000),
                                                      [Calibration.Calibration(-0.5, 1 / 1000),
                                                       Calibration.Calibration(-0.5, 1 / 1000)])

    def test_copy_graphic(self):
        rect_graphic = Graphics.RectangleGraphic()
        copy.deepcopy(rect_graphic)
        ellipse_graphic = Graphics.EllipseGraphic()
        copy.deepcopy(ellipse_graphic)
        line_graphic = Graphics.LineGraphic()
        copy.deepcopy(line_graphic)

    def test_rect_test(self):
        mapping = self.__get_mapping()
        rect_graphic = Graphics.RectangleGraphic()
        rect_graphic.bounds = (0.25, 0.25), (0.5, 0.5)
        ui_settings = DisplayPanel.FixedUISettings()
        self.assertEqual(rect_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((500, 500)), move_only=False), ("all", False))
        self.assertEqual(rect_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((250, 250)), move_only=False)[0], "top-left")
        self.assertEqual(rect_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((750, 750)), move_only=False)[0], "bottom-right")
        self.assertEqual(rect_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((250, 750)), move_only=False)[0], "top-right")
        self.assertEqual(rect_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((750, 250)), move_only=False)[0], "bottom-left")
        self.assertEqual(rect_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((250, 500)), move_only=False), ("all", True))
        self.assertEqual(rect_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((750, 500)), move_only=False), ("all", True))
        self.assertEqual(rect_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((500, 250)), move_only=False), ("all", True))
        self.assertEqual(rect_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((500, 750)), move_only=False), ("all", True))
        self.assertIsNone(rect_graphic.test(mapping, ui_settings, Geometry.FloatPoint(), move_only=False)[0])

    def test_line_test(self):
        mapping = self.__get_mapping()
        line_graphic = Graphics.LineGraphic()
        line_graphic.start = (0.25,0.25)
        line_graphic.end = (0.75,0.75)
        ui_settings = DisplayPanel.FixedUISettings()
        self.assertEqual(line_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((500, 500)), move_only=True)[0], "all")
        self.assertEqual(line_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((250, 250)), move_only=True)[0], "start")
        self.assertEqual(line_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((750, 750)), move_only=True)[0], "end")
        self.assertEqual(line_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((250, 250)), move_only=False)[0], "start")
        self.assertEqual(line_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((750, 750)), move_only=False)[0], "end")
        self.assertIsNone(line_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((240, 240)), move_only=False)[0])
        self.assertIsNone(line_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((760, 760)), move_only=False)[0])
        self.assertIsNone(line_graphic.test(mapping, ui_settings, Geometry.FloatPoint(), move_only=False)[0])

    def test_point_test(self):
        mapping = self.__get_mapping()
        point_graphic = Graphics.PointGraphic()
        point_graphic.position = (0.25,0.25)
        ui_settings = DisplayPanel.FixedUISettings()
        self.assertEqual(point_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((250, 250)), move_only=True)[0], "all")
        self.assertEqual(point_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((250 - 18, 250)), move_only=True)[0], None)
        point_graphic.label = "Test"
        self.assertEqual(point_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((250 - 18 - 6, 250)), move_only=True)[0], "all")

    def test_spot_test(self):
        mapping = self.__get_mapping()
        spot_graphic = Graphics.SpotGraphic()
        spot_graphic.bounds = (0.2 - 0.5, 0.2 - 0.5), (0.1, 0.1)
        ui_settings = DisplayPanel.FixedUISettings()
        self.assertEqual(spot_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((200, 200)), move_only=False), ("top-left", True))
        self.assertEqual(spot_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((300, 200)), move_only=False), ("bottom-left", True))
        self.assertEqual(spot_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((300, 300)), move_only=False), ("bottom-right", True))
        self.assertEqual(spot_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((200, 300)), move_only=False), ("top-right", True))
        self.assertEqual(spot_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((800, 800)), move_only=False), ("inverted-top-left", True))
        self.assertEqual(spot_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((700, 800)), move_only=False), ("inverted-bottom-left", True))
        self.assertEqual(spot_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((700, 700)), move_only=False), ("inverted-bottom-right", True))
        self.assertEqual(spot_graphic.test(mapping, ui_settings, Geometry.FloatPoint.make((800, 700)), move_only=False), ("inverted-top-right", True))
        self.assertIsNone(spot_graphic.test(mapping, ui_settings, Geometry.FloatPoint(), move_only=False)[0])

    def test_create_all_graphic_by_dragging(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            t = [
                ("line", Graphics.LineGraphic),
                ("rectangle", Graphics.RectangleGraphic),
                ("ellipse", Graphics.EllipseGraphic),
                ("point", Graphics.PointGraphic),
                ("spot", Graphics.SpotGraphic),
                ("wedge", Graphics.WedgeGraphic),
                ("ring", Graphics.RingGraphic),
                ("lattice", Graphics.LatticeGraphic),
            ]
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            for tool_mode, graphic_type in t:
                document_controller.tool_mode = tool_mode
                self.assertEqual(0, len(display_item.graphics))
                display_panel.display_canvas_item.simulate_drag((500, 500), (600, 600))
                self.assertEqual(1, len(display_item.graphics))
                graphic = display_item.graphics[0]
                self.assertIsInstance(graphic, graphic_type)
                display_item.remove_graphic(graphic)

    def test_spot_mask_is_sensible_when_smaller_than_one_pixel(self):
        spot_graphic = Graphics.SpotGraphic()
        spot_graphic.bounds = (0.0, 0.0), (0.05, 0.05)
        mask_data = spot_graphic.get_mask((10, 10), Geometry.FloatPoint(y=5.5, x=5.5))
        self.assertEqual(mask_data.shape, (10, 10))
        self.assertTrue(numpy.array_equal(mask_data, numpy.zeros((10, 10))))

    def test_spot_mask_is_sensible_when_outside_bounds(self):
        spot_graphic = Graphics.SpotGraphic()
        spot_graphic.bounds = (1.5, 1.5), (0.1, 0.1)
        mask_data = spot_graphic.get_mask((10, 10))
        self.assertEqual(mask_data.shape, (10, 10))
        self.assertTrue(numpy.array_equal(mask_data, numpy.zeros((10, 10))))
        mask_data = spot_graphic.get_mask((10, 10))
        spot_graphic.bounds = (-0.5, -0.5), (0.1, 0.1)
        self.assertEqual(mask_data.shape, (10, 10))
        self.assertTrue(numpy.array_equal(mask_data, numpy.zeros((10, 10))))

    def test_spot_mask_is_sensible_when_partially_outside_bounds(self):
        spot_graphic = Graphics.SpotGraphic()
        spot_graphic.bounds = (0.25, 0.25), (0.5, 0.5)
        mask_data = spot_graphic.get_mask((10, 10))
        self.assertEqual(mask_data.shape, (10, 10))
        self.assertFalse(numpy.array_equal(mask_data, numpy.zeros((10, 10))))
        spot_graphic.bounds = (-0.75, -0.75), (0.5, 0.5)
        mask_data = spot_graphic.get_mask((10, 10))
        self.assertEqual(mask_data.shape, (10, 10))
        self.assertFalse(numpy.array_equal(mask_data, numpy.zeros((10, 10))))

    def assertAlmostEqualPoint(self, p1, p2, e=0.00001):
        if not(Geometry.distance(p1, p2) < e):
            logging.debug("%s != %s", p1, p2)
        self.assertTrue(Geometry.distance(p1, p2) < e)

    def assertAlmostEqualSize(self, s1, s2, e=0.00001):
        if not(abs(s2.height - s1.height) < e) or not(abs(s2.width - s1.width) < e):
            logging.debug("%s != %s", s1, s2)
        self.assertTrue(abs(s2.height - s1.height) < e and abs(s2.width - s1.width) < e)

    def assertAlmostEqualRect(self, r1, r2, e=0.00001):
        if not(Geometry.distance(r1[0], r2[0]) < e and Geometry.distance(r1[1], r2[1]) < e):
            logging.debug("%s != %s", r1, r2)
        self.assertTrue(Geometry.distance(r1[0], r2[0]) < e and Geometry.distance(r1[1], r2[1]) < e)

    def test_dragging_regions(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller_with_application()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))

            def get_extended_attr(object, extended_name):
                initial_value = object
                for sub_property in extended_name.split("."):
                    initial_value = getattr(initial_value, sub_property)
                return initial_value

            def map_string(s, m1, m2):
                for k, v in m1.items():
                    s = s.replace(k, v)
                for k, v in m2.items():
                    s = s.replace(k, v)
                return s

            class ScalarCoordinate(object):

                def __init__(self, v_value=None, h_value=None):
                    self.v_value = v_value
                    self.h_value = h_value

                @property
                def value(self):
                    return self.v_value if self.v_value is not None else self.h_value

            def reflect_rect(r, v, h):
                if v and not h:
                    return Geometry.FloatRect(Geometry.FloatPoint(1.0 - r.bottom, r.left), r.size)
                elif v and h:
                    return Geometry.FloatRect(Geometry.FloatPoint(1.0 - r.bottom, 1.0 - r.right), r.size)
                elif not v and h:
                    return Geometry.FloatRect(Geometry.FloatPoint(r.top, 1.0 - r.right), r.size)
                else:
                    return r

            def reflect_point(p, v, h):
                y = 1.0 - p.y if v else p.y
                x = 1.0 - p.x if h else p.x
                return Geometry.FloatPoint(y, x)

            def reflect_size(s, v, h):
                height = 1.0 - s.height if v else s.height
                width = 1.0 - s.width if h else s.width
                return Geometry.FloatSize(height, width)

            def reflect_scalar(s, v, h):
                v_value = 1.0 - s.v_value if s.v_value is not None and v else s.v_value
                h_value = 1.0 - s.h_value if s.h_value is not None and h else s.h_value
                return ScalarCoordinate(v_value, h_value)

            def reflect_value(value, v, h, mapping_v1, mapping_v2, mapping_h1, mapping_h2):
                if isinstance(value, Geometry.FloatRect):
                    return reflect_rect(value, v, h)
                elif isinstance(value, Geometry.FloatPoint):
                    return reflect_point(value, v, h)
                elif isinstance(value, Geometry.FloatSize):
                    return reflect_size(value, v, h)
                elif isinstance(value, ScalarCoordinate):
                    return reflect_scalar(value, v, h)
                elif isinstance(value, str):
                    if v:
                        value = map_string(value, mapping_v1, mapping_v2)
                    if h:
                        value = map_string(value, mapping_h1, mapping_h2)
                    return value
                elif isinstance(value, float):
                    return value
                raise Exception("Unknown value type %s", type(value))
                return None

            def reflect(d, v, h):
                d = copy.deepcopy(d)
                mapping_v1 = {"top": "was_top", "bottom": "was_bottom"}
                mapping_v2 = {"was_top": "bottom", "was_bottom": "top"}
                mapping_h1 = {"left": "was_left", "right": "was_right"}
                mapping_h2 = {"was_left": "right", "was_right": "left"}
                # name
                if v:
                    d["name"] = map_string(d["name"], mapping_v1, mapping_v2)
                if h:
                    d["name"] = map_string(d["name"], mapping_h1, mapping_h2)
                # input
                old_input = d["input"]["properties"]
                new_input = dict()
                for property, expected_value in old_input.items():
                    if v:
                        property = map_string(property, mapping_v1, mapping_v2)
                    if h:
                        property = map_string(property, mapping_h1, mapping_h2)
                    new_input[property] = reflect_value(expected_value, v, h, mapping_v1, mapping_v2, mapping_h1, mapping_h2)
                d["input"]["properties"] = new_input
                # drag
                drag = d["drag"]
                c0, c1 = drag[0:2]
                if v:
                    c0 = 1000 - c0[0], c0[1]
                    c1 = 1000 - c1[0], c1[1]
                if h:
                    c0 = c0[0], 1000 - c0[1]
                    c1 = c1[0], 1000 - c1[1]
                new_drag = list((c0, c1))
                new_drag.extend(drag[2:])
                d["drag"] = tuple(new_drag)
                # output
                old_output = d["output"]["properties"]
                new_output = dict()
                for property, expected_value in old_output.items():
                    if v:
                        property = map_string(property, mapping_v1, mapping_v2)
                    if h:
                        property = map_string(property, mapping_h1, mapping_h2)
                    new_output[property] = reflect_value(expected_value, v, h, mapping_v1, mapping_v2, mapping_h1, mapping_h2)
                d["output"]["properties"] = new_output
                return d

            def do_drag_test(d, e=0.00001):
                # logging.debug("test %s", d["name"])
                region = Graphics.factory(lambda t: d["input"]["type"])
                for property, initial_value in d["input"]["properties"].items():
                    setattr(region, property, initial_value)
                initial_values = dict()
                for property, expected_value in d["output"]["properties"].items():
                    if isinstance(expected_value, str):
                        initial_values[expected_value] = get_extended_attr(region, expected_value)
                for constraint in d["input"].get("constraints", list()):
                    if constraint == "bounds":
                        region.is_bounds_constrained = True
                    elif constraint == "shape":
                        region.is_shape_locked = True
                    elif constraint == "position":
                        region.is_position_locked = True
                display_item.add_graphic(region)
                display_item.graphic_selection.set(0)
                display_panel.display_canvas_item.simulate_drag(*d["drag"])
                for property, expected_value in d["output"]["properties"].items():
                    actual_value = get_extended_attr(region, property)
                    # logging.debug("%s: %s == %s ?", property, actual_value, expected_value)
                    if isinstance(expected_value, str):
                        if isinstance(actual_value, Geometry.FloatRect):
                            self.assertAlmostEqualRect(actual_value, initial_values[expected_value], e)
                        elif isinstance(actual_value, Geometry.FloatPoint):
                            self.assertAlmostEqualPoint(actual_value, initial_values[expected_value], e)
                        elif isinstance(actual_value, Geometry.FloatSize):
                            self.assertAlmostEqualSize(actual_value, initial_values[expected_value], e)
                        elif isinstance(actual_value, ScalarCoordinate):
                            self.assertAlmostEqual(actual_value.value, initial_values[expected_value])
                        else:
                            raise Exception("Unknown value type %s", type(actual_value))
                    else:
                        if isinstance(actual_value, Geometry.FloatRect):
                            self.assertAlmostEqualRect(actual_value, expected_value, e)
                        elif isinstance(actual_value, Geometry.FloatPoint):
                            self.assertAlmostEqualPoint(actual_value, expected_value, e)
                        elif isinstance(actual_value, Geometry.FloatSize):
                            self.assertAlmostEqualSize(actual_value, expected_value, e)
                        elif isinstance(actual_value, float):
                            self.assertAlmostEqual(actual_value, expected_value.value)
                        else:
                            raise Exception("Unknown value type %s", type(actual_value))
                display_item.remove_graphic(region)

            def rotate(p, o, angle):
                origin = Geometry.FloatPoint.make(o)
                point = Geometry.FloatPoint.make(p)
                delta = point - origin
                angle_sin = math.sin(angle)
                angle_cos = math.cos(angle)
                return origin + Geometry.FloatPoint(x=delta.x * angle_cos + delta.y * angle_sin, y=delta.y * angle_cos - delta.x * angle_sin)

            def round1000(p):
                return Geometry.FloatPoint(y=int(p.y * 1000) / 1000, x=int(p.x * 1000) / 1000)

            # rectangle top-left

            d = {
                "name": "drag top-left corner outside of bounds with no constraints",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect(origin=(0.21, 0.22), size=(0.31, 0.32)) }
                },
                "drag": [(210, 220), (-190, -180)],
                "output": {
                    "properties": {
                        "_bounds.top_left": Geometry.FloatPoint(y=-0.19, x=-0.18),
                        "_bounds.bottom_right": "_bounds.bottom_right"
                    },
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            for rotation in (2 * math.pi / 8, -2 * math.pi / 8):

                d = {
                    "name": "drag rotated top-left corner outside of bounds with no constraints",
                    "input": {
                        "type": "rect-graphic",
                        "properties": { "_bounds": Geometry.FloatRect(origin=(0.4, 0.3), size=(0.2, 0.4)), "rotation": rotation }
                    },
                    "drag": [rotate((400, 300), (500, 500), rotation), rotate((450, 250), (500, 500), rotation)],
                    "output": {
                        "properties": {
                            "_rotated_top_left": rotate((0.45, 0.25), (0.5, 0.5), rotation),
                            "_rotated_bottom_right": "_rotated_bottom_right"
                        },
                    }
                }

                # rotation odd reflections are not valid
                do_drag_test(reflect(d, False, False), e=0.001)
                do_drag_test(reflect(d, True, True), e=0.001)

                d = {
                    "name": "drag rotated top-right corner outside of bounds with no constraints",
                    "input": {
                        "type": "rect-graphic",
                        "properties": { "_bounds": Geometry.FloatRect(origin=(0.4, 0.3), size=(0.2, 0.4)), "rotation": rotation }
                    },
                    "drag": [rotate((400, 700), (500, 500), rotation), rotate((450, 750), (500, 500), rotation)],
                    "output": {
                        "properties": {
                            "_rotated_top_right": rotate((0.45, 0.75), (0.5, 0.5), rotation),
                            "_rotated_bottom_left": "_rotated_bottom_left"
                        },
                    }
                }

                # rotation odd reflections are not valid
                do_drag_test(reflect(d, False, False), e=0.001)
                do_drag_test(reflect(d, True, True), e=0.001)

                d = {
                    "name": "drag rotated top-left corner outside of bounds from center",
                    "input": {
                        "type": "rect-graphic",
                        "properties": { "_bounds": Geometry.FloatRect(origin=(0.4, 0.3), size=(0.2, 0.4)), "rotation": rotation }
                    },
                    "drag": [rotate((400, 300), (500, 500), rotation), rotate((450, 250), (500, 500), rotation), CanvasItem.KeyboardModifiers(alt=True)],
                    "output": {
                        "properties": {
                            "_rotated_top_left": rotate((0.45, 0.25), (0.5, 0.5), rotation),
                            "_bounds.center": "_bounds.center"
                        },
                    }
                }

                # rotation odd reflections are not valid
                do_drag_test(reflect(d, False, False), e=0.001)
                do_drag_test(reflect(d, True, True), e=0.001)

                d = {
                    "name": "drag rotated top-right corner outside of bounds from center",
                    "input": {
                        "type": "rect-graphic",
                        "properties": { "_bounds": Geometry.FloatRect(origin=(0.4, 0.3), size=(0.2, 0.4)), "rotation": rotation }
                    },
                    "drag": [rotate((400, 700), (500, 500), rotation), rotate((450, 750), (500, 500), rotation), CanvasItem.KeyboardModifiers(alt=True)],
                    "output": {
                        "properties": {
                            "_rotated_top_right": rotate((0.45, 0.75), (0.5, 0.5), rotation),
                            "_bounds.center": "_bounds.center"
                        },
                    }
                }

                # rotation odd reflections are not valid
                do_drag_test(reflect(d, False, False), e=0.001)
                do_drag_test(reflect(d, True, True), e=0.001)

            d = {
                "name": "drag top-left corner outside of bounds with bounds constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.21, 0.22), (0.31, 0.32)) },
                    "constraints": ["bounds"]
                },
                "drag": [(210, 220), (-190, -180)],
                "output": {
                    "properties": {
                        "_bounds.top_left": Geometry.FloatPoint(),
                        "_bounds.bottom_right": "_bounds.bottom_right"
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag top-left corner from center towards top left with bounds constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.25, 0.25), (0.5, 0.5)) },
                    "constraints": ["bounds"]
                },
                "drag": [(250, 250), (150, 140), CanvasItem.KeyboardModifiers(alt=True)],
                "output": {
                    "properties": {
                        "_bounds.center": "_bounds.center",
                        "_bounds.top_left": Geometry.FloatPoint(0.15, 0.14),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag top-left corner from center towards top left with position constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.25, 0.25), (0.5, 0.5)) },
                    "constraints": ["position"]
                },
                "drag": [(250, 250), (150, 140), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_bounds.center": "_bounds.center",
                        "_bounds.top_left": Geometry.FloatPoint(0.15, 0.14),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag top-left corner from center outside of bounds to the top left with bounds constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.21, 0.22), (0.31, 0.32)) },
                    "constraints": ["bounds"]
                },
                "drag": [(210, 220), (-190, -180), CanvasItem.KeyboardModifiers(alt=True)],
                "output": {
                    "properties": {
                        "_bounds.center": "_bounds.center",
                        "_bounds.top_left": Geometry.FloatPoint(),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag top-left corner from center outside of bounds to the bottom with bounds constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.21, 0.22), (0.31, 0.32)) },
                    "constraints": ["bounds"]
                },
                "drag": [(210, 220), (1500, 220), CanvasItem.KeyboardModifiers(alt=True)],
                "output": {
                    "properties": {
                        "_bounds.center": "_bounds.center",
                        "_bounds.top": ScalarCoordinate(v_value=0.0),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag top-left corner from center outside of bounds to the right with bounds constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.21, 0.22), (0.31, 0.32)) },
                    "constraints": ["bounds"]
                },
                "drag": [(210, 220), (210, 1500), CanvasItem.KeyboardModifiers(alt=True)],
                "output": {
                    "properties": {
                        "_bounds.center": "_bounds.center",
                        "_bounds.left": ScalarCoordinate(h_value=0.0),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag squared top-left corner from outside of bounds to the bottom right with bounds constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.2, 0.3), (0.4, 0.5)) },
                    "constraints": ["bounds"]
                },
                "drag": [(200, 300), (1500, 1500), CanvasItem.KeyboardModifiers(shift=True)],
                "output": {
                    "properties": {
                        "_bounds.top_left": "_bounds.bottom_right",
                        "_bounds.bottom": ScalarCoordinate(v_value=0.8),
                        "_bounds.right": ScalarCoordinate(h_value=1.0),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag squared top-left corner from in one direction",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.4, 0.4), (0.2, 0.2)) },  # center 0.5, 0.5
                    "constraints": ["bounds"]
                },
                "drag": [(400, 400), (200, 400), CanvasItem.KeyboardModifiers(shift=True)],
                "output": {
                    "properties": {
                        "_bounds.bottom_right": "_bounds.bottom_right",
                        "_bounds.top_left": Geometry.FloatPoint(0.2, 0.2),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag squared top-left corner from outside of bounds to the bottom right with bounds constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.1, 0.3), (0.2, 0.2)) },  # center 0.2, 0.4
                    "constraints": ["bounds"]
                },
                "drag": [(100, 300), (-200, -200), CanvasItem.KeyboardModifiers(shift=True, alt=True)],
                "output": {
                    "properties": {
                        "_bounds.center": "_bounds.center",
                        "_bounds.left": ScalarCoordinate(h_value=0.2),
                        "_bounds.top": ScalarCoordinate(v_value=0.0),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag squared top-left corner with bounds, center, square constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.4, 0.4), (0.2, 0.2)) },  # center 0.5, 0.5
                    "constraints": ["bounds"]
                },
                "drag": [(400, 400), (300, 200), CanvasItem.KeyboardModifiers(shift=True, alt=True)],
                "output": {
                    "properties": {
                        "_bounds.center": "_bounds.center",
                        "_bounds.top_left": Geometry.FloatPoint(0.3, 0.3),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag squared top-left corner with bounds, center, square constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.7, 0.7), (0.2, 0.2)) },  # center 0.8, 0.8
                    "constraints": ["bounds"]
                },
                "drag": [(700, 700), (0, 0), CanvasItem.KeyboardModifiers(shift=True, alt=True)],
                "output": {
                    "properties": {
                        "_bounds.center": "_bounds.center",
                        "_bounds.right": ScalarCoordinate(h_value=1.0),
                        "_bounds.bottom": ScalarCoordinate(v_value=1.0),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag top-left to top-left with shape constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.3, 0.2), (0.2, 0.2)) },  # center 0.4, 0.3
                    "constraints": ["shape"]
                },
                "drag": [(300, 200), (-100, -100), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_bounds.center": Geometry.FloatPoint(),
                        "_bounds.size": "_bounds.size",
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag all to top-left with no constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.3, 0.2), (0.2, 0.2)) },  # center 0.4, 0.3
                },
                "drag": [(400, 300), (0, 0), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_bounds.center": Geometry.FloatPoint(),
                        "_bounds.size": "_bounds.size",
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag all top-left with bounds constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.3, 0.2), (0.2, 0.2)) },  # center 0.4, 0.3
                    "constraints": ["bounds"]
                },
                "drag": [(400, 300), (0, 0), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_bounds.center": Geometry.FloatPoint(0.1, 0.1),
                        "_bounds.size": "_bounds.size",
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag all bottom-right with bounds constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.3, 0.2), (0.2, 0.2)) },  # center 0.4, 0.3
                    "constraints": ["bounds"]
                },
                "drag": [(400, 300), (1000, 1000), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_bounds.center": Geometry.FloatPoint(0.9, 0.9),
                        "_bounds.size": "_bounds.size",
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "drag all with restrict constraint",
                "input": {
                    "type": "rect-graphic",
                    "properties": { "_bounds": Geometry.FloatRect((0.4, 0.4), (0.2, 0.2)) },  # center 0.5, 0.5
                },
                "drag": [(500, 500), (800, 600), CanvasItem.KeyboardModifiers(shift=True)],
                "output": {
                    "properties": {
                        "_bounds.center": Geometry.FloatPoint(0.8, 0.5),
                        "_bounds.size": "_bounds.size",
                    }
                }
            }

            # vertical reflections not valid
            for h in (False, True):
                do_drag_test(reflect(d, False, h))

            # point

            d = {
                "name": "point drag with no constraint",
                "input": {
                    "type": "point-graphic",
                    "properties": { "_position": Geometry.FloatPoint(0.2, 0.3) },
                    # "constraints": ["bounds"]
                },
                "drag": [(200, 300), (-100, -100), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_position": Geometry.FloatPoint(-0.1, -0.1),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "point drag with bounds constraint",
                "input": {
                    "type": "point-graphic",
                    "properties": { "_position": Geometry.FloatPoint(0.2, 0.3) },
                    "constraints": ["bounds"]
                },
                "drag": [(200, 300), (-100, -100), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_position": Geometry.FloatPoint(),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "point drag with restrict",
                "input": {
                    "type": "point-graphic",
                    "properties": { "_position": Geometry.FloatPoint(0.2, 0.3) },
                    "constraints": ["bounds"]
                },
                "drag": [(200, 300), (100, 100), CanvasItem.KeyboardModifiers(shift=True)],
                "output": {
                    "properties": {
                        "_position": Geometry.FloatPoint(0.2, 0.1),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            # line regions

            d = {
                "name": "line drag with no constraints",
                "input": {
                    "type": "line-graphic",
                    "properties": {
                        "_start": Geometry.FloatPoint(0.2, 0.3),
                        "_end": Geometry.FloatPoint(0.6, 0.5),
                    },
                    # "constraints": ["bounds"]
                },
                "drag": [(400, 400), (600, 700), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_start": Geometry.FloatPoint(0.4, 0.6),
                        "_end": Geometry.FloatPoint(0.8, 0.8),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "line start drag with restrict",
                "input": {
                    "type": "line-graphic",
                    "properties": {
                        "_start": Geometry.FloatPoint(0.2, 0.3),
                        "_end": Geometry.FloatPoint(0.6, 0.5),
                    },
                    "constraints": ["shape"]
                },
                "drag": [(200, 300), (400, 600), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_start": Geometry.FloatPoint(0.4, 0.6),
                        "_end": Geometry.FloatPoint(0.8, 0.8),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "line drag with bounds constraint",
                "input": {
                    "type": "line-graphic",
                    "properties": {
                        "_start": Geometry.FloatPoint(0.2, 0.3),
                        "_end": Geometry.FloatPoint(0.6, 0.5),
                    },
                    "constraints": ["bounds"]
                },
                "drag": [(400, 400), (1000, 1000), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_start": Geometry.FloatPoint(0.6, 0.8),
                        "_end": Geometry.FloatPoint(1.0, 1.0),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "line start drag with no constraints",
                "input": {
                    "type": "line-graphic",
                    "properties": {
                        "_start": Geometry.FloatPoint(0.2, 0.3),
                        "_end": Geometry.FloatPoint(0.6, 0.5),
                    },
                    # "constraints": ["bounds"]
                },
                "drag": [(200, 300), (600, 700), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_start": Geometry.FloatPoint(0.6, 0.7),
                        "_end": Geometry.FloatPoint(0.6, 0.5),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "line start drag with bounds constraint",
                "input": {
                    "type": "line-graphic",
                    "properties": {
                        "_start": Geometry.FloatPoint(0.2, 0.3),
                        "_end": Geometry.FloatPoint(0.6, 0.5),
                    },
                    "constraints": ["bounds"]
                },
                "drag": [(200, 300), (-100, 300), CanvasItem.KeyboardModifiers()],
                "output": {
                    "properties": {
                        "_start": Geometry.FloatPoint(0.0, 0.3),
                        "_end": Geometry.FloatPoint(0.6, 0.5),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

            d = {
                "name": "line start drag with shape constraint",
                "input": {
                    "type": "line-graphic",
                    "properties": {
                        "_start": Geometry.FloatPoint(0.2, 0.3),
                        "_end": Geometry.FloatPoint(0.6, 0.5),
                    },
                    "constraints": ["bounds"]
                },
                "drag": [(200, 300), (500, 500), CanvasItem.KeyboardModifiers(shift=True)],
                "output": {
                    "properties": {
                        "_start": Geometry.FloatPoint(0.5, 0.5),
                        "_end": Geometry.FloatPoint(0.6, 0.5),
                    }
                }
            }

            for v in (False, True):
                for h in (False, True):
                    do_drag_test(reflect(d, v, h))

    def test_selected_graphics_get_priority_when_dragging_middles(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            rect_graphic1 = Graphics.RectangleGraphic()
            rect_graphic2 = Graphics.RectangleGraphic()
            rect_graphic1.bounds = (0.4, 0.4), (0.2, 0.2)
            rect_graphic2.bounds = (0.25, 0.25), (0.5, 0.5)
            display_item.add_graphic(rect_graphic1)
            display_item.add_graphic(rect_graphic2)
            display_item.graphic_selection.set(0)
            # now the smaller rectangle (selected) is behind the larger one
            display_panel.display_canvas_item.simulate_drag((500, 500), (600, 600))
            # make sure the smaller one gets dragged
            self.assertAlmostEqualPoint(Geometry.FloatRect.make(rect_graphic1.bounds).center, Geometry.FloatPoint(y=0.6, x=0.6))
            self.assertAlmostEqualPoint(Geometry.FloatRect.make(rect_graphic2.bounds).center, Geometry.FloatPoint(y=0.5, x=0.5))

    def test_removing_graphic_from_display_closes_it(self):
        # make the document controller
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            graphic = Graphics.PointGraphic()
            display_item.add_graphic(graphic)
            self.assertFalse(graphic._closed)
            display_item.remove_graphic(graphic)
            self.assertTrue(graphic._closed)

    def test_removing_data_item_closes_graphic_attached_to_display(self):
        # make the document controller
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            graphic = Graphics.PointGraphic()
            display_item.add_graphic(graphic)
            self.assertFalse(graphic._closed)
            document_model.remove_data_item(data_item)
            self.assertTrue(graphic._closed)

    def test_removing_region_closes_associated_drawn_graphic(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            point_region = Graphics.PointGraphic()
            display_item.add_graphic(point_region)
            drawn_graphic = display_item.graphics[0]
            self.assertFalse(drawn_graphic._closed)
            display_item.remove_graphic(point_region)
            self.assertTrue(drawn_graphic._closed)

    def test_removing_data_item_closes_associated_drawn_graphic(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            point_region = Graphics.PointGraphic()
            display_item.add_graphic(point_region)
            drawn_graphic = display_item.graphics[0]
            self.assertFalse(drawn_graphic._closed)
            document_model.remove_data_item(data_item)
            self.assertTrue(drawn_graphic._closed)

    def test_removing_graphic_with_dependent_data_removes_dependent_data(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            crop_region = Graphics.RectangleGraphic()
            display_item.add_graphic(crop_region)
            cropped_data_item = document_model.get_crop_new(display_item, display_item.data_item, crop_region)
            self.assertIn(cropped_data_item, document_model.data_items)
            display_item.remove_graphic(crop_region)
            self.assertNotIn(cropped_data_item, document_model.data_items)

    def test_removing_one_of_two_graphics_with_dependent_data_only_removes_the_one(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            crop_region1 = Graphics.RectangleGraphic()
            display_item.add_graphic(crop_region1)
            crop_region2 = Graphics.RectangleGraphic()
            display_item.add_graphic(crop_region2)
            document_model.get_crop_new(display_item, display_item.data_item, crop_region1)
            document_model.get_crop_new(display_item, display_item.data_item, crop_region2)
            self.assertIn(crop_region1, display_item.graphics)
            self.assertIn(crop_region2, display_item.graphics)
            display_item.remove_graphic(crop_region1)
            self.assertIn(crop_region2, display_item.graphics)

    def test_changing_data_length_does_not_update_graphics(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item = DataItem.DataItem(numpy.zeros((100, ), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            interval_graphic = Graphics.IntervalGraphic()
            interval_graphic.interval = (0.25, 0.75)
            display_item.add_graphic(interval_graphic)
            self.assertEqual(display_item.displayed_dimensional_scales[-1], 100)
            data_item.set_data(numpy.zeros((50, ), numpy.uint32))
            self.assertEqual(interval_graphic.interval, (0.25, 0.75))

    def test_changing_data_scale_does_not_update_graphics(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item = DataItem.DataItem(numpy.zeros((100, ), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            interval_graphic = Graphics.IntervalGraphic()
            interval_graphic.interval = (0.25, 0.75)
            display_item.add_graphic(interval_graphic)
            self.assertEqual(display_item.displayed_dimensional_scales[-1], 100)
            display_item.dimensional_scales = (1,)
            self.assertEqual(interval_graphic.interval, (0.25, 0.75))

    def test_setting_interval_to_non_floats_throws_execption(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item = DataItem.DataItem(numpy.zeros((100, ), numpy.uint32))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            interval_graphic = Graphics.IntervalGraphic()
            interval_graphic.interval = (0.25, 0.75)
            display_item.add_graphic(interval_graphic)
            with self.assertRaises(Exception):
                interval_graphic.interval = (numpy.array([1, 2]), numpy.array([3, 4]))
            with self.assertRaises(Exception):
                interval_graphic.start = numpy.array([1, 2])
            with self.assertRaises(Exception):
                interval_graphic.end = numpy.array([3, 4])


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



from __future__ import annotations

import abc
import copy
import json
import os
import pathlib
import typing

from nion.utils import Converter
from nion.utils import Observable


ANY = "any"
STRING = "string"
BOOLEAN = "boolean"
INT = "integer"
FLOAT = "double"
LIST = "list"
DICT = "dict"
SET = "set"
TIMESTAMP = "timestamp"
UUID = "uuid"
PATH = "path"

OPTIONAL = True
REQUIRED = False

entity_types: typing.Dict[str, EntityType] = dict()


def register_entity_type(entity_id: str, entity: EntityType) -> None:
    entity_types[entity_id] = entity

def get_entity_type(entity_id: str) -> EntityType:
    return entity_types.get(entity_id)

def build_value(type: str, value) -> typing.Any:
    if value is None:
        return None
    if type in (SET, ):
        return set(value)
    if type in (BOOLEAN, ):
        return bool(value)
    if type in (INT, ):
        return int(value)
    if type in (FLOAT, ):
        return float(value)
    if type in (TIMESTAMP):
        return Converter.DatetimeToStringConverter().convert_back(value)
    if type in (UUID):
        return Converter.UuidToStringConverter().convert_back(value)
    if type in (PATH):
        return Converter.PathToStringConverter().convert_back(value)
    return value


def dict_value(type: str, value):
    if value is None:
        return None
    if type in (SET, ):
        return list(value)
    if type in (BOOLEAN, ):
        return bool(value)
    if type in (INT, ):
        return int(value)
    if type in (FLOAT, ):
        return float(value)
    if type in (TIMESTAMP):
        return Converter.DatetimeToStringConverter().convert(value)
    if type in (UUID):
        return Converter.UuidToStringConverter().convert(value)
    if type in (PATH):
        return Converter.PathToStringConverter().convert(value)
    return value


DictValue = typing.Optional[typing.Union[typing.Dict, typing.List, typing.Tuple, str]]


class Field(abc.ABC):
    def close(self) -> None:
        pass

    @abc.abstractmethod
    def read(self, dict_value: typing.Any) -> Field: ...

    @abc.abstractmethod
    def write(self) -> DictValue: ...

    @property
    @abc.abstractmethod
    def field_value(self) -> typing.Any: ...

    @field_value.setter
    @abc.abstractmethod
    def field_value(self, value: typing.Any) -> None: ...


class PropertyField(Field):
    def __init__(self, type: str, optional: bool, default):
        self.__type = type
        self.__optional = optional
        self.__default = default
        self.__value = default

    def read(self, dict_value: typing.Any) -> Field:
        self.__value = build_value(self.__type, dict_value)
        return self

    def write(self) -> DictValue:
        return dict_value(self.__type, self.__value)

    @property
    def field_value(self) -> typing.Any:
        return self.__value

    @field_value.setter
    def field_value(self, value: typing.Any) -> None:
        self.__value = value


class TupleField(Field):
    def __init__(self, type: FieldType, optional: bool, default_values):
        assert isinstance(type, FieldType)
        assert default_values is None or isinstance(default_values, (tuple, list))
        self.__type = type
        self.__optional = optional
        self.__default_values = list(default_values) if default_values is not None else None
        self.__fields = None
        self.read(default_values)

    def close(self) -> None:
        self.__clear()
        super().close()

    def __clear(self):
        if self.__fields is not None:
            for field in self.__fields:
                field.close()
            self.__fields = None

    def read(self, dict_value: typing.Any) -> Field:
        self.__clear()
        if isinstance(dict_value, (tuple, list)):
            self.__fields = tuple(self.__type.create(v) for v in dict_value)
        return self

    def write(self) -> DictValue:
        if self.__fields is not None:
            return list(field.write() for field in self.__fields)
        else:
            return None if self.__optional else copy.copy(self.__default_values)

    @property
    def field_value(self) -> typing.Any:
        return [field.field_value for field in self.__fields]

    @field_value.setter
    def field_value(self, value: typing.Any) -> None:
        assert isinstance(value, (tuple, list))
        self.__fields = tuple(self.__type.create(v) for v in value)


class FixedTupleField(Field):
    def __init__(self, types: typing.Sequence[FieldType], optional: bool, default_values):
        assert default_values is None or isinstance(default_values, (tuple, list))
        self.__types = types
        self.__optional = optional
        self.__default_values = list(default_values) if default_values is not None else None
        self.__fields = None
        self.read(default_values)

    def close(self) -> None:
        self.__clear()
        super().close()

    def __clear(self):
        if self.__fields is not None:
            for field in self.__fields:
                field.close()
            self.__fields = None

    def read(self, dict_value: typing.Any) -> Field:
        # TODO: the fields should be fixed
        self.__clear()
        if isinstance(dict_value, (tuple, list)):
            self.__fields = tuple(type.create(v) for type, v in zip(self.__types, dict_value))
        else:
            self.__value = tuple()
        return self

    def write(self) -> DictValue:
        if self.__fields is not None:
            return list(field.write() for field in self.__fields)
        else:
            return None if self.__optional else copy.copy(self.__default_values)

    @property
    def field_value(self) -> typing.Any:
        return [field.field_value for field in self.__fields]

    @field_value.setter
    def field_value(self, value: typing.Any) -> None:
        assert isinstance(value, (tuple, list))
        self.__fields = tuple(type.create(v) for type, v in zip(self.__types, value))


class RecordField(Field):
    def __init__(self, field_type_map: typing.Mapping[str, FieldType]):
        self.__field_type_map = field_type_map
        self.__field_map = dict()

    def close(self) -> None:
        for field in self.__field_map.values():
            field.close()
        self.__field_map = None
        super().close()

    def read(self, dict_value: typing.Any) -> Field:
        if isinstance(dict_value, (dict)):
            self.__field_map = {k: type.create(dict_value.get(k)) for k, type in self.__field_type_map.items()}
        else:
            self.__value = tuple()
        return self

    def write(self) -> DictValue:
        d = dict()
        for k, field in self.__field_map.items():
            v = field.write()
            if v is not None:
                d[k] = v
        return d

    @property
    def field_value(self) -> typing.Any:
        return None

    @field_value.setter
    def field_value(self, value: typing.Any) -> None:
        pass


class ArrayField(Field):
    def __init__(self, type: FieldType, optional: bool):
        self.__type = type
        self.__optional = optional
        self.__fields = list()

    def read(self, dict_value: typing.Any) -> Field:
        if isinstance(dict_value, (tuple, list)):
            self.__fields = list(self.__type.create(item) for item in dict_value)
        else:
            self.__fields = list()
        return self

    def write(self) -> DictValue:
        if self.__fields is not None:
            l = list(field.write() for field in self.__fields)
        else:
            l = list()
        return l if l or not self.__optional else None

    @property
    def field_value(self) -> typing.Any:
        return [field.field_value for field in self.__fields]

    @field_value.setter
    def field_value(self, value: typing.Any) -> None:
        assert isinstance(value, (tuple, list))
        self.__fields = list(self.__type.create(v) for v in value)


class MapField(Field):
    def __init__(self, key: FieldType, value: FieldType, optional: bool):
        self.__key = key
        self.__value = value
        self.__optional = optional
        self.__map = dict()

    def read(self, dict_value: typing.Any) -> Field:
        if isinstance(dict_value, dict):
            self.__map = {k: self.__value.create(v) for k, v in dict_value.items()}
        else:
            self.__map = dict()
        return self

    def write(self) -> DictValue:
        if self.__map is not None:
            d = {k: v.write() for k, v in self.__map.items()}
        else:
            d = dict()
        return d if d or not self.__optional else None

    @property
    def field_value(self) -> typing.Any:
        return {k: field.field_value for k, field in self.__map}

    @field_value.setter
    def field_value(self, value: typing.Any) -> None:
        assert isinstance(value, dict)
        self.__fields = {k: self.__value.create(v) for k, v in value.items()}


class ReferenceField(Field):
    def __init__(self, type: EntityType):
        self.__type = type
        self.__reference = None

    def read(self, dict_value: typing.Any) -> Field:
        self.__reference = copy.deepcopy(dict_value)
        return self

    def write(self) -> DictValue:
        return copy.deepcopy(self.__reference)

    @property
    def field_value(self) -> typing.Any:
        return None

    @field_value.setter
    def field_value(self, value: typing.Any) -> None:
        pass


class ComponentField(Field):
    def __init__(self, entity_id: str):
        self.__type = get_entity_type(entity_id)
        self.__field = None

    def close(self) -> None:
        if self.__field:
            self.__field.close()
            self.__field = None
        super().close()

    def read(self, dict_value: typing.Any) -> Field:
        self.__field = self.__type.create(dict_value)
        return self

    def write(self) -> DictValue:
        return self.__field.write_to_dict()

    @property
    def field_value(self) -> typing.Any:
        return None

    @field_value.setter
    def field_value(self, value: typing.Any) -> None:
        pass


class FieldType(abc.ABC):
    def __init__(self, field_class: typing.Callable[..., Field], *args, **kwargs):
        assert callable(field_class)
        self.__field_class = field_class
        self.__args = args
        self.__kwargs = kwargs

    @property
    def _field_class(self) -> typing.Callable[..., Field]:
        return self.__field_class

    @property
    def _args(self):
        return self.__args

    @property
    def _kwargs(self):
        return self.__kwargs

    def _call(self, field_class: typing.Callable[..., Field], *args, **kwargs) -> Field:
        return field_class(*args, **kwargs)

    def create(self, dict_value: DictValue) -> Field:
        return self._call(self.__field_class, *self.__args, **self.__kwargs).read(dict_value)


class PropertyType(FieldType):
    def __init__(self, type: str, optional: bool, default):
        super().__init__(PropertyField, type, optional, default)
        self.type = type
        self.optional = optional
        self.default = default


class TupleType(FieldType):
    def __init__(self, type: FieldType, optional: bool, default):
        super().__init__(TupleField, type, optional, default)
        self.type = type
        self.optional = optional


class FixedTupleType(FieldType):
    def __init__(self, types: typing.Sequence[FieldType], optional: bool, default):
        super().__init__(FixedTupleField, types, optional, default)
        self.types = list(types)
        self.optional = optional


class RecordType(FieldType):
    def __init__(self, field_type_map: typing.Mapping[str, FieldType]):
        super().__init__(RecordField, field_type_map)
        self.__field_type_map = dict(field_type_map)


class ArrayType(FieldType):
    def __init__(self, type: FieldType, optional: bool):
        super().__init__(ArrayField, type, optional)
        self.type = type
        self.optional = optional


class MapType(FieldType):
    def __init__(self, key: FieldType, value: FieldType, optional: bool):
        super().__init__(MapField, key, value, optional)
        self.key = key
        self.value = value
        self.optional = optional


class ReferenceType(FieldType):
    def __init__(self, type: EntityType):
        super().__init__(ReferenceField, type)
        self.type = type


class ComponentType(FieldType):
    def __init__(self, entity_id: str):
        super().__init__(ComponentField, entity_id)
        self.entity_id = entity_id

    def create(self, dict_value: DictValue) -> Field:
        assert isinstance(dict_value, dict)
        d = dict_value
        entity_type = get_entity_type(d.get("type"))
        field_type = entity_type.entity_id if entity_type else self._args[0]
        return self._call(self._field_class, *((field_type, ) + self._args[1:]), **self._kwargs).read(dict_value)


class Entity(Observable.Observable):
    def __init__(self, type: EntityType, version: typing.Optional[int], field_type_map: typing.Mapping[str, FieldType], renames: typing.Mapping[str, str]):
        super().__init__()
        self.__entity_type = type
        self.__version = version
        self.__field_type_map = field_type_map
        self.__field_dict : typing.Dict[str, Field] = dict()
        self.__renames = renames
        self.writer: typing.Optional[typing.Callable[[typing.Dict], None]] = None

    def close(self) -> None:
        pass

    def read(self, properties: typing.Mapping) -> Entity:
        self.__field_dict = dict()
        for field_name, field_type in self.__field_type_map.items():
            d = properties.get(self.__renames.get(field_name, field_name))
            self.__field_dict[field_name] = field_type.create(d)
        return self

    def write_to_dict(self) -> typing.Dict:
        d = dict()
        d["type"] = self.__entity_type.entity_id
        if self.__version:
            d["version"] = self.__version
        for field_name, field in self.__field_dict.items():
            dd = field.write()
            if dd is not None:
                d[self.__renames.get(field_name, field_name)] = dd
        return d

    @property
    def entity_type(self) -> EntityType:
        return self.__entity_type

    def __getattr__(self, name: str) -> typing.Any:
        field_dict = self.__dict__.get("_Entity__field_dict", dict())
        if name in field_dict:
            return field_dict[name].field_value

    def __setattr__(self, name: str, value: typing.Any) -> None:
        field_dict = self.__dict__.get("_Entity__field_dict", dict())
        if name in field_dict:
            self.__set_field_value(name, value)
        else:
            super().__setattr__(name, value)

    def __set_field_value(self, name: str, value: typing.Any) -> None:
        self.__field_dict[name].field_value = value
        self.property_changed_event.fire(name)
        if callable(self.writer):
            self.writer(self.write_to_dict())


class EntityType:
    def __init__(self, entity_id: str, base: typing.Optional[EntityType], version: typing.Optional[int], field_type_map: typing.Mapping[str, FieldType]):
        self.__entity_id = entity_id
        self.__base = base
        self.__version = version
        self.__renames = dict()
        self.__field_type_map = dict()
        self.__field_type_map["uuid"] = prop(UUID)
        self.__field_type_map["modified"] = prop(TIMESTAMP)
        while base:
            for k, v in base._field_type_map.items():
                if not k in ("uuid", "modified"):
                    self.__field_type_map[k] = v
            for k, v in base._renames.items():
                self.__renames[k] = v
            base = base.base
        for k, v in field_type_map.items():
            self.__field_type_map[k] = v
        register_entity_type(entity_id, self)

    @property
    def _field_type_map(self) -> typing.Mapping[str, FieldType]:
        return self.__field_type_map

    @property
    def _renames(self) -> typing.Mapping[str, str]:
        return self.__renames

    def create(self, d: typing.Dict) -> Entity:
        return Entity(self, self.__version, self.__field_type_map, self.__renames).read(d)

    @property
    def base(self) -> typing.Optional[EntityType]:
        return self.__base

    @property
    def entity_id(self) -> str:
        return self.__entity_id

    def rename(self, field_name: str, storage_field_name: str) -> None:
        self.__renames[field_name] = storage_field_name

    def is_subclass_of(self, entity_type: EntityType) -> bool:
        return self.base and (self.base == entity_type or self.base.is_subclass_of(entity_type))

    @property
    def subclasses(self) -> typing.List[EntityType]:
        return [entity_type for entity_type in entity_types.values() if entity_type.is_subclass_of(self)]

    @property
    def concrete_subclasses(self) -> typing.List[EntityType]:
        return [entity_type for entity_type in entity_types.values() if entity_type.is_subclass_of(self) and not entity_type.subclasses]


def prop(type: str, optional: bool=False, *, default=None) -> PropertyType:
    return PropertyType(type, optional, default)

def indefinite_tuple(type: FieldType, optional: bool=False, default=None) -> TupleType:
    return TupleType(type, optional, default)

def fixed_tuple(types: typing.List[FieldType], optional: bool = False, default=None) -> FixedTupleType:
    return FixedTupleType(types, optional, default)

def record(field_type_map: typing.Dict[str, FieldType]) -> RecordType:
    return RecordType(field_type_map)

def array(type: FieldType, optional: bool = False) -> ArrayType:
    return ArrayType(type, optional)

def map(key: FieldType, value: FieldType, optional: bool = False) -> MapType:
    return MapType(key, value, optional)

def reference(type: typing.Optional[EntityType] = None) -> ReferenceType:
    return ReferenceType(type)

def component(type: typing.Union[EntityType, str]) -> ComponentType:
    if isinstance(type, EntityType):
        return ComponentType(type.entity_id)
    else:
        return ComponentType(type)

def entity(entity_id: str, base: typing.Optional[EntityType], version: typing.Optional[int], field_type_map: typing.Dict[str, FieldType]) -> EntityType:
    return EntityType(entity_id, base, version, field_type_map)

def read_json(path: pathlib.Path) -> typing.Dict:
    """Read the json file from path. Rename path if not readable."""
    properties = dict()
    if path and path.exists():
        try:
            with path.open("r") as fp:
                properties = json.load(fp)
        except Exception:
            os.replace(path, path.with_suffix(".bak"))
    return properties

def read_entity(entity_type: EntityType, path: pathlib.Path, o: str) -> Entity:
    """Read the entity from the path."""
    properties = read_json(path)
    return entity_type.create(properties)



# standard libraries
import copy
import typing

# third party libraries

# local libraries
from nion.swift.model import Changes
from nion.swift.model import DataItem
from nion.swift.model import DisplayItem
from nion.swift.model import Graphics
from nion.swift.model import Persistence
from nion.swift.model import Schema
from nion.utils import Event
from nion.utils import Observable

if typing.TYPE_CHECKING:
    from nion.swift.model import Project


class DataStructure(Observable.Observable, Persistence.PersistentObject):
    entity_types:typing.Dict[str, Schema.EntityType] = dict()
    entity_names:typing.Dict[str, str] = dict()
    entity_package_names:typing.Dict[str, str] = dict()

    # regarding naming: https://en.wikipedia.org/wiki/Passive_data_structure
    def __init__(self, *, structure_type: str=None, source=None):
        super().__init__()
        self.__properties = dict()
        self.__referenced_object_proxies = dict()
        self.define_type("data_structure")
        self.define_property("structure_type", structure_type, changed=self.__structure_type_changed)
        self.define_property("source_specifier", changed=self.__source_specifier_changed, key="source_uuid")
        # properties is handled explicitly
        self.data_structure_changed_event = Event.Event()
        self.referenced_objects_changed_event = Event.Event()
        self.__source_proxy = self.create_item_proxy(item=source)
        self.source_specifier = source.project.create_specifier(source).write() if source else None
        self.__entity: typing.Optional[Schema.Entity] = None
        self.__create_entity()

    def close(self) -> None:
        if self.__entity:
            self.__entity.close()
            self.__entity = None
        self.__source_proxy.close()
        self.__source_proxy = None
        for referenced_proxy in self.__referenced_object_proxies.values():
            referenced_proxy.close()
        self.__referenced_object_proxies.clear()
        super().close()

    @classmethod
    def register_entity(cls, entity_type: Schema.EntityType, *, entity_name: str = None, entity_package_name: str = None, **kwargs) -> None:
        DataStructure.entity_types[entity_type.entity_id] = entity_type
        if entity_name:
            DataStructure.entity_names[entity_type.entity_id] = entity_name
        if entity_package_name:
            DataStructure.entity_package_names[entity_type.entity_id] = entity_package_name

    @classmethod
    def unregister_entity(cls, structure_type: str) -> None:
        DataStructure.entity_types.pop(structure_type)
        DataStructure.entity_names.pop(structure_type, None)
        DataStructure.entity_package_names.pop(structure_type, None)

    def __getattr__(self, name):
        properties = self.__dict__.get("_DataStructure__properties", dict())
        if name in properties:
            return properties[name]
        return super().__getattr__(name)

    def __setattr__(self, name, value):
        properties = self.__dict__.get("_DataStructure__properties", dict())
        if name in properties:
            if value is not None:
                self.set_property_value(name, value)
            else:
                self.remove_property_value(name)
        else:
            super().__setattr__(name, value)

    @property
    def project(self) -> "Project.Project":
        return typing.cast("Project.Project", self.container)

    def create_proxy(self) -> Persistence.PersistentObjectProxy:
        return self.project.create_item_proxy(item=self)

    @property
    def item_specifier(self) -> Persistence.PersistentObjectSpecifier:
        return Persistence.PersistentObjectSpecifier(item_uuid=self.uuid, context_uuid=self.project.uuid)

    def insert_model_item(self, container, name, before_index, item):
        if self.container:
            self.container.insert_model_item(container, name, before_index, item)
        else:
            container.insert_item(name, before_index, item)

    def remove_model_item(self, container, name, item, *, safe: bool=False) -> Changes.UndeleteLog:
        if self.container:
            return self.container.remove_model_item(container, name, item, safe=safe)
        else:
            container.remove_item(name, item)
            return Changes.UndeleteLog()

    def read_from_dict(self, properties):
        super().read_from_dict(properties)
        self.__properties = properties.get("properties")
        for property_name, value in self.__properties.items():
            self.__configure_reference_proxy(property_name, value, None)
        self.__create_entity()

    def __create_entity(self) -> None:
        if self.__entity:
            self.__entity.close()
            self.__entity = None

        if self.structure_type in DataStructure.entity_types:
            self.__entity = DataStructure.entity_types[self.structure_type].create(self.__properties)

            def write_entity(d: typing.Dict) -> None:
                for property, value in d.items():
                    if property != "type":
                        self.__properties[property] = value
                        reference_object_proxy = self.__referenced_object_proxies.pop(property, None)
                        if reference_object_proxy:
                            reference_object_proxy.close()
                        self.__configure_reference_proxy(property, value, None)
                        self.data_structure_changed_event.fire(property)
                        self.property_changed_event.fire(property)
                self._update_persistent_property("properties", self.__properties)

            self.__entity.writer = write_entity

    def __configure_reference_proxy(self, property_name, value, item):
        if isinstance(value, dict) and value.get("type") in {"data_item", "display_item", "data_source", "graphic", "structure"} and "uuid" in value:
            self.__referenced_object_proxies[property_name] = self.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(value["uuid"]), item=item)

    def write_to_dict(self):
        properties = super().write_to_dict()
        properties["properties"] = copy.deepcopy(self.__properties)
        return properties

    @property
    def source(self):
        return self.__source_proxy.item

    @source.setter
    def source(self, source):
        self.__source_proxy.item = source
        self.source_specifier = source.project.create_specifier(source).write() if source else None

    @property
    def entity(self) -> typing.Optional[Schema.Entity]:
        return self.__entity

    def __structure_type_changed(self, name: str, structure_type: str) -> None:
        self.__create_entity()
        self.property_changed_event.fire("structure_type")

    def __source_specifier_changed(self, name: str, d: typing.Dict) -> None:
        self.__source_proxy.item_specifier = Persistence.PersistentObjectSpecifier.read(d)

    def set_property_value(self, property: str, value) -> None:
        self.__properties[property] = value
        reference_object_proxy = self.__referenced_object_proxies.pop(property, None)
        if reference_object_proxy:
            reference_object_proxy.close()
        self.__configure_reference_proxy(property, value, None)
        self.data_structure_changed_event.fire(property)
        self.property_changed_event.fire(property)
        self._update_persistent_property("properties", self.__properties)

    def remove_property_value(self, property: str) -> None:
        if property in self.__properties:
            self.__properties.pop(property)
            reference_object_proxy = self.__referenced_object_proxies.pop(property, None)
            if reference_object_proxy:
                reference_object_proxy.close()
            self.data_structure_changed_event.fire(property)
            self.property_changed_event.fire(property)
            self._update_persistent_property("properties", self.__properties)

    def get_property_value(self, property: str, default_value=None):
        return self.__properties.get(property, default_value)

    def set_referenced_object(self, property: str, item) -> None:
        assert item is not None
        if item != self.get_referenced_object(property):
            object_type = "data_item" if isinstance(item, DataItem.DataItem) else None
            self.__properties[property] = get_object_specifier(item, object_type)
            reference_object_proxy = self.__referenced_object_proxies.pop(property, None)
            if reference_object_proxy:
                reference_object_proxy.close()
            self.__configure_reference_proxy(property, self.__properties[property], item)
            self.data_structure_changed_event.fire(property)
            self.property_changed_event.fire(property)
            self._update_persistent_property("properties", self.__properties)
            self.referenced_objects_changed_event.fire()

    def remove_referenced_object(self, property: str) -> None:
        self.remove_property_value(property)

    def get_referenced_object(self, property: str):
        return self.__referenced_object_proxies[property].item if property in self.__referenced_object_proxies else None

    @property
    def referenced_objects(self) -> typing.List:
        return list(referenced_object_proxy.item for referenced_object_proxy in self.__referenced_object_proxies.values())


def get_object_specifier(object, object_type: str = None, project=None, *, allow_partial: bool = True) -> typing.Optional[typing.Dict]:
    # project is passed for testing only
    if isinstance(object, DataItem.DataItem):
        project = project or object.project
        specifier = project.create_specifier(object, allow_partial=allow_partial) if project else None
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": object_type or "data_item", "uuid": str(specifier_uuid)}
        if specifier and specifier.context_uuid: d["context_uuid"] = str(specifier.context_uuid)
        return d
    if object and object_type in ("xdata", "display_xdata", "cropped_xdata", "cropped_display_xdata", "filter_xdata", "filtered_xdata"):
        assert isinstance(object, DisplayItem.DisplayDataChannel)
        project = project or object.project
        specifier = project.create_specifier(object, allow_partial=allow_partial) if project else None
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": object_type, "uuid": str(specifier_uuid)}
        if specifier and specifier.context_uuid: d["context_uuid"] = str(specifier.context_uuid)
        return d
    if isinstance(object, DisplayItem.DisplayDataChannel):
        # should be "data_source" but requires file format change
        project = project or object.project
        specifier = project.create_specifier(object, allow_partial=allow_partial) if project else None
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": "data_source", "uuid": str(specifier_uuid)}
        if specifier and specifier.context_uuid: d["context_uuid"] = str(specifier.context_uuid)
        return d
    elif isinstance(object, Graphics.Graphic):
        project = project or object.project
        specifier = project.create_specifier(object, allow_partial=allow_partial) if project else None
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": "graphic", "uuid": str(specifier_uuid)}
        if specifier and specifier.context_uuid: d["context_uuid"] = str(specifier.context_uuid)
        return d
    elif isinstance(object, DataStructure):
        project = project or object.project
        specifier = project.create_specifier(object, allow_partial=allow_partial) if project else None
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": "structure", "uuid": str(specifier_uuid)}
        if specifier and specifier.context_uuid: d["context_uuid"] = str(specifier.context_uuid)
        return d
    elif isinstance(object, DisplayItem.DisplayItem):
        project = project or object.project
        specifier = project.create_specifier(object, allow_partial=allow_partial) if project else None
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": "display_item", "uuid": str(specifier_uuid)}
        if specifier and specifier.context_uuid: d["context_uuid"] = str(specifier.context_uuid)
        return d
    return None

ElementalMappingEdge = Schema.entity("elemental_mapping_edge", None, None, {
    "atomic_number": Schema.prop(Schema.INT),
    "shell_number": Schema.prop(Schema.INT),
    "subshell_index": Schema.prop(Schema.INT),
    "fit_interval": Schema.fixed_tuple([Schema.prop(Schema.FLOAT), Schema.prop(Schema.FLOAT)]),
    "signal_interval": Schema.fixed_tuple([Schema.prop(Schema.FLOAT), Schema.prop(Schema.FLOAT)]),
})

DataStructure.register_entity(ElementalMappingEdge, entity_name="Elemental Mapping Edge")



"""
    Provide symbolic math services.

    The goal is to provide a module (namespace) where users can be provided with variables representing
    data items (directly or indirectly via reference to workspace panels).

    DataNodes represent data items, operations, numpy arrays, and constants.
"""

# standard libraries
import ast
import contextlib
import copy
import functools
import threading
import time
import typing
import uuid
import weakref

# third party libraries
import numpy

# local libraries
from nion.data import Core
from nion.data import DataAndMetadata
from nion.swift.model import Changes
from nion.swift.model import DataItem
from nion.swift.model import DataStructure
from nion.swift.model import DisplayItem
from nion.swift.model import Graphics
from nion.swift.model import Persistence
from nion.utils import Converter
from nion.utils import Event
from nion.utils import Geometry
from nion.utils import ListModel
from nion.utils import Observable

if typing.TYPE_CHECKING:
    from nion.swift.model import Project


class ComputationVariableType:
    """Defines a type of a computation variable beyond the built-in types."""
    def __init__(self, type_id: str, label: str, object_type):
        self.type_id = type_id
        self.label = label
        self.object_type = object_type
        self.__objects = dict()  # type: typing.Dict[uuid.UUID, typing.Any]

    def get_object_by_uuid(self, object_uuid: uuid.UUID):
        return self.__objects.get(object_uuid)

    def register_object(self, object):
        assert object.uuid not in self.__objects
        self.__objects[object.uuid] = object

    def unregister_object(self, object):
        assert object.uuid in self.__objects
        del self.__objects[object.uuid]


class ComputationOutput(Observable.Observable, Persistence.PersistentObject):
    """Tracks an output of a computation."""

    def __init__(self, name: str=None, specifier: dict=None, specifiers: typing.Sequence[dict]=None, label: str=None):  # defaults are None for factory
        super().__init__()
        self.define_type("output")
        self.define_property("name", name, changed=self.__property_changed)
        self.define_property("label", label if label else name, hidden=True, changed=self.__property_changed)
        self.define_property("specifier", specifier, changed=self.__property_changed)
        self.define_property("specifiers", specifiers, changed=self.__property_changed)
        self.needs_rebind_event = Event.Event()  # an event to be fired when the computation needs to rebind
        self.bound_item = None
        self.__needs_rebind_event_listeners = list()

    def close(self):
        # TODO: this is not called
        for needs_rebind_event_listener in self.__needs_rebind_event_listeners:
            needs_rebind_event_listener.close()
        self.__needs_rebind_event_listeners = None
        super().close()

    def __property_changed(self, name, value):
        self.notify_property_changed(name)
        if name in ("specifier", "specifiers"):
            self.needs_rebind_event.fire()

    def __unbind(self):
        # self.specifier = None
        self.bound_item = None
        self.notify_property_changed("bound_item")

    def bind(self, resolve_object_specifier_fn):
        if self.specifier:
            self.bound_item = resolve_object_specifier_fn(self.specifier)
            if self.bound_item:
                self.__needs_rebind_event_listeners.append(self.bound_item.needs_rebind_event.listen(self.__unbind))
        elif self.specifiers is not None:
            bound_items = [resolve_object_specifier_fn(specifier) for specifier in self.specifiers]
            bound_items = [bound_item for bound_item in bound_items if bound_item is not None]
            for bound_item in bound_items:
                self.__needs_rebind_event_listeners.append(bound_item.needs_rebind_event.listen(self.__unbind))
            self.bound_item = bound_items
        else:
            self.bound_item = None
        self.notify_property_changed("bound_item")

    @property
    def output_items(self) -> typing.Set:
        item = self.bound_item
        if isinstance(item, list):
            return {list_item.value for list_item in item}
        elif item:
            return {item.value}
        return set()

    @property
    def label(self) -> str:
        # for registered computations, the computation class takes precedence in defining the labels.
        computation = self.container
        if isinstance(computation, Computation):
            compute_class = _computation_types.get(computation.processing_id)
            if compute_class:
                label = getattr(compute_class, "outputs", dict()).get(self.name, dict()).get("label")
                if label:
                    return label
        # not a registered computation, fall back to label or name.
        return self._get_persistent_property_value("label") or self.name

    @label.setter
    def label(self, value: str) -> None:
        self._set_persistent_property_value("label", value)


class ComputationVariable(Observable.Observable, Persistence.PersistentObject):
    """Tracks a variable (value or object) used in a computation.

    A variable has user visible name, a label used in the script, a value type.

    Scalar value types have a value, a default, and optional min and max values. The control type is used to
    specify the preferred UI control (e.g. checkbox vs. input field).

    Specifier value types have a specifier/secondary_specifier/property_name which can be resolved to a part of a
    specific object. The specifier indicates the object and the part of the object to be used (e.g., a data item and the
    masked data of that data item). The secondary specifier is used to augment the first object (e.g. a crop graphic on
    an image). The property name is also used to augment the specifier (e.g., a field of a data structure or graphic).

    The object provides four events: changed, fired when anything changes; variable_type_changed, fired when the
    variable type changes; needs_rebind, fired when a specifier changes and the variable needs rebinding to the context;
    and needs_rebuild, fired when the UI needs rebuilding. variable_type_changed and needs_rebuild are specific to the
    inspector and shouldn't be used elsewhere.

    Clients can ask for the bound_variable which supplies an object that provides a read-only value property and a
    changed_event. This object can be used to watch for changes to the value type portion of this object.

    Clients can also get/set the bound_item, which must be an object that provides a read-only value property and a
    changed_event.  This object can be used to watch for changes to the object portion of this object.
    """
    def __init__(self, name: str=None, *, property_name: str=None, value_type: str=None, value=None, value_default=None, value_min=None, value_max=None, control_type: str=None, specifier: dict=None, label: str=None, secondary_specifier: dict=None, items: typing.List["ComputationItem"]=None):  # defaults are None for factory
        super().__init__()
        self.define_type("variable")
        self.define_property("name", name, changed=self.__property_changed)
        self.define_property("label", label if label else name, changed=self.__property_changed)
        self.define_property("value_type", value_type, changed=self.__property_changed)
        self.define_property("value", value, changed=self.__property_changed, reader=self.__value_reader, writer=self.__value_writer)
        self.define_property("value_default", value_default, changed=self.__property_changed, reader=self.__value_reader, writer=self.__value_writer)
        self.define_property("value_min", value_min, changed=self.__property_changed, reader=self.__value_reader, writer=self.__value_writer)
        self.define_property("value_max", value_max, changed=self.__property_changed, reader=self.__value_reader, writer=self.__value_writer)
        self.define_property("specifier", specifier, changed=self.__property_changed)
        self.define_property("secondary_specifier", secondary_specifier, changed=self.__property_changed)
        self.define_property("property_name", property_name, changed=self.__property_changed)
        self.define_property("control_type", control_type, changed=self.__property_changed)
        item_specifiers = [DataStructure.get_object_specifier(item.item, item.type, allow_partial=False) if item else None for item in items] if items is not None else None
        self.define_property("object_specifiers", copy.deepcopy(item_specifiers) if item_specifiers is not None else None, changed=self.__property_changed)
        self.changed_event = Event.Event()
        self.variable_type_changed_event = Event.Event()
        self.needs_rebind_event = Event.Event()  # an event to be fired when the computation needs to rebind
        self.needs_rebuild_event = Event.Event()  # an event to be fired when the UI needs a rebuild
        self.__bound_items_model = None
        self.__bound_items_model_item_inserted_event_listener = None
        self.__bound_items_model_item_removed_event_listener = None
        self.__bound_item = None
        self.__bound_item_changed_event_listener = None
        self.__bound_item_removed_event_listener = None
        self.__bound_item_child_removed_event_listener = None

    def close(self):
        # TODO: this is not called
        if self.__bound_items_model_item_inserted_event_listener:
            self.__bound_items_model_item_inserted_event_listener.close()
            self.__bound_items_model_item_inserted_event_listener = None
        if self.__bound_items_model_item_removed_event_listener:
            self.__bound_items_model_item_removed_event_listener.close()
            self.__bound_items_model_item_removed_event_listener = None
        super().close()

    def __repr__(self):
        return "{} ({} {} {} {} {})".format(super().__repr__(), self.name, self.label, self.value, self.specifier, self.secondary_specifier)

    def read_from_dict(self, properties: dict) -> None:
        # used for persistence
        # ensure that value_type is read first
        value_type_property = self._get_persistent_property("value_type")
        value_type_property.read_from_dict(properties)
        super().read_from_dict(properties)

    def write_to_dict(self) -> dict:
        # used for persistence. left here since read_from_dict is defined.
        return super().write_to_dict()

    def connect_items(self, bound_items: typing.List) -> None:
        self.disconnect_items()

        bound_items_model = ListModel.ListModel(items=bound_items) if bound_items is not None else None
        self.__bound_items_model = bound_items_model
        self.__bound_items_model_item_inserted_event_listener = None
        self.__bound_items_model_item_removed_event_listener = None
        if bound_items_model is not None:

            def item_inserted(key, value, index):
                # self.changed_event.fire()  # implicit when setting object_specifiers
                object_specifiers = self.object_specifiers
                object_specifiers.insert(index, copy.deepcopy(value.specifier))
                self.object_specifiers = object_specifiers
                self.bound_item.item_inserted(index, self.__bound_items_model.items[index])
                self.needs_rebind_event.fire()

            def item_removed(key, value, index):
                # self.changed_event.fire()  # implicit when setting object_specifiers
                object_specifiers = self.object_specifiers
                object_specifiers.pop(index)
                self.object_specifiers = object_specifiers
                self.bound_item.item_removed(index)
                self.needs_rebind_event.fire()

            self.__bound_items_model_item_inserted_event_listener = self.__bound_items_model.item_inserted_event.listen(item_inserted)
            self.__bound_items_model_item_removed_event_listener = self.__bound_items_model.item_removed_event.listen(item_removed)

    def disconnect_items(self) -> None:
        if self.__bound_items_model:
            self.__bound_items_model.close()
            self.__bound_items_model = None
            self.__bound_items_model_item_inserted_event_listener.close()
            self.__bound_items_model_item_inserted_event_listener = None
            self.__bound_items_model_item_removed_event_listener.close()
            self.__bound_items_model_item_removed_event_listener = None

    def save_properties(self):
        # used for undo
        return self.value, self.specifier, self.secondary_specifier

    def restore_properties(self, properties):
        # used for undo
        self.value = properties[0]
        self.specifier = properties[1]
        self.secondary_specifier = properties[2]

    def __value_reader(self, persistent_property, properties):
        value_type = self.value_type
        raw_value = properties.get(persistent_property.key)
        if raw_value is not None:
            if value_type == "boolean":
                return bool(raw_value)
            elif value_type == "integral":
                return int(raw_value)
            elif value_type == "real":
                return float(raw_value)
            elif value_type == "complex":
                return complex(*raw_value)
            elif value_type == "string":
                return str(raw_value)
        return None

    def __value_writer(self, persistent_property, properties, value):
        value_type = self.value_type
        if value is not None:
            if value_type == "boolean":
                properties[persistent_property.key] = bool(value)
            if value_type == "integral":
                properties[persistent_property.key] = int(value)
            if value_type == "real":
                properties[persistent_property.key] = float(value)
            if value_type == "complex":
                properties[persistent_property.key] = complex(value).real, complex(value).imag
            if value_type == "string":
                properties[persistent_property.key] = str(value)

    @property
    def variable_specifier(self) -> dict:
        """Return the variable specifier for this variable.

        The specifier can be used to lookup the value of this variable in a computation context.
        """
        if self.value_type is not None:
            return {"type": "variable", "version": 1, "uuid": str(self.uuid), "x-name": self.name, "x-value": self.value}
        else:
            return self.specifier

    @property
    def bound_variable(self) -> "BoundItemBase":
        """Return an object with a value property and a changed_event.

        The value property returns the value of the variable. The changed_event is fired
        whenever the value changes.
        """

        class BoundVariable(BoundItemBase):
            def __init__(self, variable):
                super().__init__(None)
                self.valid = True
                self.__variable = variable

                def property_changed(key):
                    if key == "value":
                        self.changed_event.fire()

                self.__variable_property_changed_listener = variable.property_changed_event.listen(property_changed)

            @property
            def value(self):
                return self.__variable.value

            @property
            def base_objects(self):
                return set()

            def close(self):
                self.__variable_property_changed_listener.close()
                self.__variable_property_changed_listener = None
                super().close()

        return BoundVariable(self)

    @property
    def bound_item(self):
        return self.__bound_item

    @bound_item.setter
    def bound_item(self, bound_item):
        if self.__bound_item_changed_event_listener:
            self.__bound_item_changed_event_listener.close()
            self.__bound_item_changed_event_listener = None
        if self.__bound_item_removed_event_listener:
            self.__bound_item_removed_event_listener.close()
            self.__bound_item_removed_event_listener = None
        if self.__bound_item_child_removed_event_listener:
            self.__bound_item_child_removed_event_listener.close()
            self.__bound_item_child_removed_event_listener = None
        if self.__bound_item:
            self.__bound_item.close()
        self.__bound_item = bound_item
        self.notify_property_changed("bound_item")
        if self.__bound_item:
            self.__bound_item_changed_event_listener = self.__bound_item.changed_event.listen(self.changed_event.fire)
            self.__bound_item_removed_event_listener = self.__bound_item.needs_rebind_event.listen(self.needs_rebind_event.fire)

    @property
    def bound_items_model(self) -> typing.Optional[ListModel.ListModel]:
        return self.__bound_items_model

    @property
    def input_items(self) -> typing.Set:
        return self.bound_item.base_objects if self.bound_item else set()

    @property
    def direct_input_items(self) -> typing.Set:
        return self.bound_item.base_objects if self.bound_item and not getattr(self.bound_item, "is_list", False) else set()

    def __property_changed(self, name, value):
        # rebind first, so that property changed listeners get the right value
        if name in ("specifier", "secondary_specifier"):
            self.needs_rebind_event.fire()
        # send the primary property changed event
        self.notify_property_changed(name)
        # now send out dependent property changed events
        if name in ["name", "label"]:
            self.notify_property_changed("display_label")
        if name in ("specifier"):
            self.notify_property_changed("specifier_uuid_str")
        if name in ("secondary_specifier"):
            self.notify_property_changed("secondary_specifier_uuid_str")
        # send out the changed event
        self.changed_event.fire()
        # finally send out the rebuild event for the inspectors
        if name in ["value_type", "value_min", "value_max", "control_type"]:
            self.needs_rebuild_event.fire()

    def notify_property_changed(self, key):
        # whenever a property changed event is fired, also fire the changed_event
        # is there a test for this? not that I can find.
        super().notify_property_changed(key)
        if key not in ("bound_item",):
            self.changed_event.fire()

    def control_type_default(self, value_type: str) -> None:
        mapping = {"boolean": "checkbox", "integral": "slider", "real": "field", "complex": "field", "string": "field"}
        return mapping.get(value_type)

    @property
    def variable_type(self) -> typing.Optional[str]:
        if self.value_type is not None:
            return self.value_type
        elif self.specifier is not None:
            specifier_type = self.specifier.get("type")
            specifier_property = self.specifier.get("property")
            return specifier_property or specifier_type
        return None

    data_item_types = ("data_item", "data", "display_data")  # used for backward compatibility

    @variable_type.setter
    def variable_type(self, value_type: str) -> None:
        if value_type != self.variable_type:
            if value_type in ("boolean", "integral", "real", "complex", "string"):
                self.specifier = None
                self.secondary_specifier = None
                self.value_type = value_type
                self.control_type = self.control_type_default(value_type)
                if value_type == "boolean":
                    self.value_default = True
                elif value_type == "integral":
                    self.value_default = 0
                elif value_type == "real":
                    self.value_default = 0.0
                elif value_type == "complex":
                    self.value_default = 0 + 0j
                else:
                    self.value_default = None
                self.value_min = None
                self.value_max = None
            elif value_type in ComputationVariable.data_item_types:
                self.value_type = None
                self.control_type = None
                self.value_default = None
                self.value_min = None
                self.value_max = None
                specifier = self.specifier or {"version": 1}
                if not specifier.get("type") in ComputationVariable.data_item_types:
                    specifier.pop("uuid", None)
                specifier["type"] = "data_source"
                if value_type in ("data", "display_data"):
                    specifier["property"] = value_type
                else:
                    specifier.pop("property", None)
                self.specifier = specifier
                self.secondary_specifier = self.secondary_specifier or {"version": 1}
            elif value_type in ("graphic"):
                self.value_type = None
                self.control_type = None
                self.value_default = None
                self.value_min = None
                self.value_max = None
                specifier = self.specifier or {"version": 1}
                specifier["type"] = value_type
                specifier.pop("uuid", None)
                specifier.pop("property", None)
                self.specifier = specifier
                self.secondary_specifier = None
            self.variable_type_changed_event.fire()

    @property
    def specifier_uuid_str(self):
        return self.specifier.get("uuid") if self.specifier else None

    @specifier_uuid_str.setter
    def specifier_uuid_str(self, value):
        converter = Converter.UuidToStringConverter()
        value = converter.convert(converter.convert_back(value))
        if self.specifier_uuid_str != value and self.specifier:
            specifier = self.specifier
            if value:
                specifier["uuid"] = value
            else:
                specifier.pop("uuid", None)
            self.specifier = specifier

    @property
    def secondary_specifier_uuid_str(self):
        return self.secondary_specifier.get("uuid") if self.secondary_specifier else None

    @secondary_specifier_uuid_str.setter
    def secondary_specifier_uuid_str(self, value):
        converter = Converter.UuidToStringConverter()
        value = converter.convert(converter.convert_back(value))
        if self.secondary_specifier_uuid_str != value and self.secondary_specifier:
            secondary_specifier = self.secondary_specifier
            if value:
                secondary_specifier["uuid"] = value
            else:
                secondary_specifier.pop("uuid", None)
            self.secondary_specifier = secondary_specifier

    @property
    def display_label(self) -> str:
        # for registered computations, the computation class takes precedence in defining the labels.
        computation = self.container
        if isinstance(computation, Computation):
            compute_class = _computation_types.get(computation.processing_id)
            if compute_class:
                label = getattr(compute_class, "inputs", dict()).get(self.name, dict()).get("label")
                if label:
                    return label
        # not a registered computation, fall back to label or name.
        return self.label or self.name

    @property
    def entity_id(self) -> typing.Optional[str]:
        computation = self.container
        if isinstance(computation, Computation):
            compute_class = _computation_types.get(computation.processing_id)
            if compute_class:
                label = getattr(compute_class, "inputs", dict()).get(self.name, dict()).get("entity_id")
                if label:
                    return label
        return None

    @property
    def has_range(self):
        return self.value_type is not None and self.value_min is not None and self.value_max is not None


def variable_factory(lookup_id):
    build_map = {
        "variable": ComputationVariable,
    }
    type = lookup_id("type")
    return build_map[type]() if type in build_map else None


def result_factory(lookup_id):
    return ComputationOutput()


class BoundItemBase:

    def __init__(self, specifier):
        self.specifier = specifier
        self.changed_event = Event.Event()
        self.needs_rebind_event = Event.Event()
        self.property_changed_event = Event.Event()
        self.valid = False

    def close(self) -> None:
        pass

    @property
    def value(self):
        return None

    @property
    def base_objects(self) -> typing.Set:
        return set()


class BoundData(BoundItemBase):

    def __init__(self, project, specifier):
        super().__init__(specifier)

        self.__item_proxy = project.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(specifier))
        self.__data_changed_event_listener = None

        def maintain_data_source():
            if self.__data_changed_event_listener:
                self.__data_changed_event_listener.close()
                self.__data_changed_event_listener = None
            object = self._object
            if object:
                self.__data_changed_event_listener = object.data_changed_event.listen(self.changed_event.fire)
            self.valid = object is not None

        def item_registered(item):
            maintain_data_source()

        def item_unregistered(item):
            maintain_data_source()
            self.needs_rebind_event.fire()

        self.__item_proxy.on_item_registered = item_registered
        self.__item_proxy.on_item_unregistered = item_unregistered

        maintain_data_source()

    def close(self):
        if self.__data_changed_event_listener:
            self.__data_changed_event_listener.close()
            self.__data_changed_event_listener = None
        self.__item_proxy.close()
        self.__item_proxy = None
        super().close()

    @property
    def base_objects(self):
        return {self._object}

    @property
    def value(self):
        return self._object.xdata

    @property
    def _object(self):
        item = self.__item_proxy.item
        if isinstance(item, DisplayItem.DisplayDataChannel):
            item = item.data_item
        return item


class BoundDisplayDataChannelBase(BoundItemBase):

    def __init__(self, project, specifier, secondary_specifier):
        super().__init__(specifier)

        self.__item_proxy = project.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(specifier))
        self.__graphic_proxy = project.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(secondary_specifier))
        self.__display_values_changed_event_listener = None

        def maintain_data_source():
            if self.__display_values_changed_event_listener:
                self.__display_values_changed_event_listener.close()
                self.__display_values_changed_event_listener = None
            display_data_channel = self._display_data_channel
            if display_data_channel:
                self.__display_values_changed_event_listener = display_data_channel.add_calculated_display_values_listener(self.changed_event.fire, send=False)
            self.valid = self.__display_values_changed_event_listener is not None

        def item_registered(item):
            maintain_data_source()

        def item_unregistered(item):
            maintain_data_source()
            self.needs_rebind_event.fire()

        self.__item_proxy.on_item_registered = item_registered
        self.__item_proxy.on_item_unregistered = item_unregistered

        self.__graphic_proxy.on_item_registered = item_registered
        self.__graphic_proxy.on_item_unregistered = item_unregistered

        maintain_data_source()

    def close(self):
        if self.__display_values_changed_event_listener:
            self.__display_values_changed_event_listener.close()
            self.__display_values_changed_event_listener = None
        self.__item_proxy.close()
        self.__item_proxy = None
        self.__graphic_proxy.close()
        self.__graphic_proxy = None
        super().close()

    @property
    def base_objects(self):
        objects = {self._display_data_channel.container, self._display_data_channel.data_item}
        if self._graphic:
            objects.add(self._graphic)
        return objects

    @property
    def _display_data_channel(self):
        return self.__item_proxy.item

    @property
    def _graphic(self):
        return self.__graphic_proxy.item


class BoundDataSource(BoundItemBase):

    def __init__(self, project, specifier, secondary_specifier):
        super().__init__(specifier)

        self.__item_proxy = project.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(specifier))
        self.__graphic_proxy = project.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(secondary_specifier))
        self.__data_source = None

        def maintain_data_source():
            if self.__data_source:
                self.__data_source.close()
                self.__data_source = None
            display_data_channel = self.__item_proxy.item
            if display_data_channel and display_data_channel.data_item:
                self.__data_source = DataItem.MonitoredDataSource(display_data_channel, self.__graphic_proxy.item, self.changed_event)
            self.valid = self.__data_source is not None

        def item_registered(item):
            maintain_data_source()

        def item_unregistered(item):
            maintain_data_source()
            self.needs_rebind_event.fire()

        self.__item_proxy.on_item_registered = item_registered
        self.__item_proxy.on_item_unregistered = item_unregistered

        self.__graphic_proxy.on_item_registered = item_registered
        self.__graphic_proxy.on_item_unregistered = item_unregistered

        maintain_data_source()

    def close(self):
        if self.__data_source:
            self.__data_source.close()
            self.__data_source = None
        self.__item_proxy.close()
        self.__item_proxy = None
        self.__graphic_proxy.close()
        self.__graphic_proxy = None
        super().close()

    @property
    def value(self):
        return self.__data_source

    @property
    def base_objects(self):
        objects = {self.__data_source.data_item}
        if self.__data_source.graphic:
            objects.add(self.__data_source.graphic)
        return objects

    @property
    def _display_data_channel(self) -> typing.Optional[DisplayItem.DisplayDataChannel]:
        return self.__item_proxy.item

    @property
    def _graphic(self) -> typing.Optional[Graphics.Graphic]:
        return self.__graphic_proxy.item


class BoundDataItem(BoundItemBase):

    def __init__(self, project, specifier):
        super().__init__(specifier)

        self.__item_proxy = project.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(specifier))
        self.__data_item_changed_event_listener = None

        def item_registered(item):
            self.__data_item_changed_event_listener = item.data_item_changed_event.listen(self.changed_event.fire)

        def item_unregistered(item):
            self.__data_item_changed_event_listener.close()
            self.__data_item_changed_event_listener = None
            self.needs_rebind_event.fire()

        self.__item_proxy.on_item_registered = item_registered
        self.__item_proxy.on_item_unregistered = item_unregistered

        if self.__item_proxy.item:
            item_registered(self.__item_proxy.item)
            self.valid = True
        else:
            self.valid = False

    def close(self):
        if self.__data_item_changed_event_listener:
            self.__data_item_changed_event_listener.close()
            self.__data_item_changed_event_listener = None
        self.__item_proxy.close()
        self.__item_proxy = None
        super().close()

    @property
    def value(self):
        return self.__item_proxy.item

    @property
    def base_objects(self):
        return {self.value}

    @property
    def _data_item(self) -> typing.Optional[DataItem.DataItem]:
        return self.__item_proxy.item


class BoundDisplayData(BoundDisplayDataChannelBase):

    @property
    def value(self):
        return self._display_data_channel.get_calculated_display_values(True).display_data_and_metadata if self._display_data_channel else None


class BoundCroppedData(BoundDisplayDataChannelBase):

    @property
    def value(self):
        xdata = self._display_data_channel.data_item.xdata
        graphic = self._graphic
        if graphic:
            if hasattr(graphic, "bounds"):
                return Core.function_crop(xdata, graphic.bounds)
            if hasattr(graphic, "interval"):
                return Core.function_crop_interval(xdata, graphic.interval)
        return xdata


class BoundCroppedDisplayData(BoundDisplayDataChannelBase):

    @property
    def value(self):
        xdata = self._display_data_channel.get_calculated_display_values(True).display_data_and_metadata
        graphic = self._graphic
        if graphic:
            if hasattr(graphic, "bounds"):
                return Core.function_crop(xdata, graphic.bounds)
            if hasattr(graphic, "interval"):
                return Core.function_crop_interval(xdata, graphic.interval)
        return xdata


class BoundFilterData(BoundDisplayDataChannelBase):

    @property
    def value(self):
        display_item = self._display_data_channel.container
        # no display item is a special case for cascade removing graphics from computations. ugh.
        # see test_new_computation_becomes_unresolved_when_xdata_input_is_removed_from_document.
        if display_item:
            shape = self._display_data_channel.get_calculated_display_values(True).display_data_and_metadata.data_shape
            calibrated_origin = Geometry.FloatPoint(y=display_item.datum_calibrations[0].convert_from_calibrated_value(0.0),
                                                    x=display_item.datum_calibrations[1].convert_from_calibrated_value(0.0))
            mask = DataItem.create_mask_data(display_item.graphics, shape, calibrated_origin)
            return DataAndMetadata.DataAndMetadata.from_data(mask)
        return None

    @property
    def base_objects(self):
        data_item = self._display_data_channel.data_item
        display_item = self._display_data_channel.container
        objects = {data_item, display_item}
        graphics = display_item.graphics if display_item else list()
        for graphic in graphics:
            if isinstance(graphic, (Graphics.PointTypeGraphic, Graphics.LineTypeGraphic, Graphics.RectangleTypeGraphic, Graphics.SpotGraphic, Graphics.WedgeGraphic, Graphics.RingGraphic, Graphics.LatticeGraphic)):
                objects.add(graphic)
        return objects


class BoundFilteredData(BoundDisplayDataChannelBase):

    @property
    def value(self):
        display_item = self._display_data_channel.container
        # no display item is a special case for cascade removing graphics from computations. ugh.
        # see test_new_computation_becomes_unresolved_when_xdata_input_is_removed_from_document.
        if display_item:
            xdata = self._display_data_channel.data_item.xdata
            if xdata.is_data_2d and xdata.is_data_complex_type:
                shape = xdata.data_shape
                calibrated_origin = Geometry.FloatPoint(y=display_item.datum_calibrations[0].convert_from_calibrated_value(0.0),
                                                        x=display_item.datum_calibrations[1].convert_from_calibrated_value(0.0))
                mask = DataItem.create_mask_data(display_item.graphics, shape, calibrated_origin)
                return Core.function_fourier_mask(xdata, DataAndMetadata.DataAndMetadata.from_data(mask))
            return xdata
        return None

    @property
    def base_objects(self):
        data_item = self._display_data_channel.data_item
        display_item = self._display_data_channel.container
        objects = {data_item, display_item}
        graphics = display_item.graphics if display_item else list()
        for graphic in graphics:
            if isinstance(graphic, (Graphics.PointTypeGraphic, Graphics.LineTypeGraphic, Graphics.RectangleTypeGraphic, Graphics.SpotGraphic, Graphics.WedgeGraphic, Graphics.RingGraphic, Graphics.LatticeGraphic)):
                objects.add(graphic)
        return objects


class BoundDataStructure(BoundItemBase):

    def __init__(self, project, specifier, property_name: str):
        super().__init__(specifier)

        self.__property_name = property_name

        self.__item_proxy = project.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(specifier))
        self.__changed_listener = None

        def data_structure_changed(property_name):
            self.changed_event.fire()
            if self.__property_name and property_name in self.__property_name:
                self.property_changed_event.fire(property_name)

        def item_registered(item):
            self.__changed_listener = item.data_structure_changed_event.listen(data_structure_changed)
            self.__property_changed_listener = item.property_changed_event.listen(data_structure_changed)

        def item_unregistered(item):
            self.__changed_listener.close()
            self.__changed_listener = None
            self.__property_changed_listener.close()
            self.__property_changed_listener = None
            self.needs_rebind_event.fire()

        self.__item_proxy.on_item_registered = item_registered
        self.__item_proxy.on_item_unregistered = item_unregistered

        if self.__item_proxy.item:
            item_registered(self.__item_proxy.item)
            self.valid = True
        else:
            self.valid = False

    def close(self):
        if self.__changed_listener:
            self.__changed_listener.close()
            self.__changed_listener = None
        self.__item_proxy.close()
        self.__item_proxy = None
        super().close()

    @property
    def value(self):
        if self.__property_name:
            return self.__object.get_property_value(self.__property_name)
        return self.__object

    @property
    def base_objects(self):
        return {self.__object}

    @property
    def __object(self):
        return self.__item_proxy.item


class BoundGraphic(BoundItemBase):

    def __init__(self, project, specifier, property_name: str):
        super().__init__(specifier)

        self.__property_name = property_name

        self.__item_proxy = project.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(specifier))
        self.__changed_listener = None

        def property_changed(property_name):
            # temporary hack to improve performance of line profile. long term solution will be to
            # have the computation decide what is a valid reason for recomputing.
            # see the test test_adjusting_interval_on_line_profile_does_not_trigger_recompute
            if not property_name in ("interval_descriptors",):
                self.changed_event.fire()
                if property_name == self.__property_name:
                    self.property_changed_event.fire(property_name)

        def item_registered(item):
            self.__changed_listener = item.property_changed_event.listen(property_changed)

        def item_unregistered(item):
            self.__changed_listener.close()
            self.__changed_listener = None
            self.needs_rebind_event.fire()

        self.__item_proxy.on_item_registered = item_registered
        self.__item_proxy.on_item_unregistered = item_unregistered

        if self.__item_proxy.item:
            item_registered(self.__item_proxy.item)
            self.valid = True
        else:
            self.valid = False

    def close(self):
        if self.__changed_listener:
            self.__changed_listener.close()
            self.__changed_listener = None
        self.__item_proxy.close()
        self.__item_proxy = None
        super().close()

    @property
    def value(self):
        if self.__property_name:
            return getattr(self.__object, self.__property_name)
        return self.__object

    @property
    def base_objects(self):
        return {self.__object}

    @property
    def __object(self):
        return self.__item_proxy.item

    @property
    def _graphic(self) -> typing.Optional[Graphics.Graphic]:
        return self.__item_proxy.item


class ComputationItem:
    def __init__(self, *, item=None, type: str=None, secondary_item=None, items: typing.List["ComputationItem"] = None):
        self.item = item
        self.type = type
        self.secondary_item = secondary_item
        self.items = items


def make_item(item, *, type: str=None, secondary_item=None) -> ComputationItem:
    return ComputationItem(item=item, type=type, secondary_item=secondary_item)


def make_item_list(items, *, type: str=None) -> ComputationItem:
    items = [make_item(item, type=type) if not isinstance(item, ComputationItem) else item for item in items]
    return ComputationItem(items=items)


class BoundList:

    def __init__(self, bound_items: typing.List):
        self.__bound_items = list()
        self.changed_event = Event.Event()
        self.needs_rebind_event = Event.Event()
        self.child_removed_event = Event.Event()
        self.property_changed_event = Event.Event()
        self.is_list = True
        self.__changed_listeners = list()
        self.__needs_rebind_listeners = list()
        self.__resolved = True
        for index, bound_item in enumerate(bound_items):
            self.item_inserted(index, bound_item)

    def close(self):
        while len(self.__bound_items) > 0:
            self.item_removed(0)
        self.__bound_items = None
        self.__resolved_items = None
        self.__changed_listeners = None

    @property
    def value(self):
        return [bound_item.value for bound_item in self.__bound_items] if self.__resolved else None

    @property
    def base_objects(self):
        # return the base objects in a stable order
        base_objects = list()
        for bound_item in self.__bound_items:
            if bound_item:
                for base_object in bound_item.base_objects:
                    if not base_object in base_objects:
                        base_objects.append(base_object)
        return base_objects

    def item_inserted(self, index, bound_item):
        self.__bound_items.insert(index, bound_item)
        self.__changed_listeners.insert(index, bound_item.changed_event.listen(self.changed_event.fire) if bound_item else None)
        self.__needs_rebind_listeners.insert(index, bound_item.needs_rebind_event.listen(functools.partial(self.child_removed_event.fire, index)) if bound_item else None)
        self.__resolved = self.__resolved and bound_item is not None

    def item_removed(self, index):
        bound_item = self.__bound_items.pop(index)
        if bound_item:
            bound_item.close()
        changed_listener = self.__changed_listeners.pop(index)
        if changed_listener:
            changed_listener.close()
        needs_rebind_listener = self.__needs_rebind_listeners.pop(index)
        if needs_rebind_listener:
            needs_rebind_listener.close()


class Computation(Observable.Observable, Persistence.PersistentObject):
    """A computation on data and other inputs.

    Watches for changes to the sources and fires a computation_mutated_event
    when a new computation needs to occur.

    Call parse_expression first to establish the computation. Bind will be automatically called.

    Call bind to establish connections after reloading. Call unbind to release connections.

    Listen to computation_mutated_event and call evaluate in response to perform
    computation (on thread).

    The computation will listen to any bound items established in the bind method. When those
    items signal a change, the computation_mutated_event will be fired.

    The processing_id is used to specify a computation that may be updated with a different script
    in the future. For instance, the line profile processing via the UI will produce a somewhat
    complicated computation expression. By recording processing_id, if the computation expression
    evolves to a better version in the future, it can be replaced with the newer version by knowing
    that the intention of the original expression was a line profile from the UI.

    The processing_id is cleared if the user changes the script expression.
    """

    def __init__(self, expression: str=None):
        super().__init__()
        self.define_type("computation")
        self.define_property("source_specifier", changed=self.__source_specifier_changed, key="source_uuid")
        self.define_property("original_expression", expression)
        self.define_property("error_text", hidden=True, changed=self.__error_changed)
        self.define_property("label", hidden=True, changed=self.__label_changed)
        self.define_property("processing_id")  # see note above
        self.define_relationship("variables", variable_factory)
        self.define_relationship("results", result_factory)
        self.__source_proxy = self.create_item_proxy()
        self.__variable_changed_event_listeners = dict()
        self.__variable_needs_rebind_event_listeners = dict()
        self.__result_needs_rebind_event_listeners = dict()
        self.last_evaluate_data_time = 0
        self.needs_update = expression is not None
        self.computation_mutated_event = Event.Event()
        self.computation_output_changed_event = Event.Event()
        self.variable_inserted_event = Event.Event()
        self.variable_removed_event = Event.Event()
        self.is_initial_computation_complete = threading.Event()  # helpful for waiting for initial computation
        self._evaluation_count_for_test = 0
        self._inputs = set()  # used by document model for tracking dependencies
        self._outputs = set()
        self.pending_project = None  # used for new computations to tell them where they'll end up
        self.__is_bound = False

    def close(self) -> None:
        self.unbind()
        self.__source_proxy.close()
        self.__source_proxy = None
        super().close()

    @property
    def project(self) -> "Project.Project":
        return typing.cast("Project.Project", self.container)

    def create_proxy(self) -> Persistence.PersistentObjectProxy:
        return self.project.create_item_proxy(item=self)

    @property
    def item_specifier(self) -> Persistence.PersistentObjectSpecifier:
        return Persistence.PersistentObjectSpecifier(item_uuid=self.uuid, context_uuid=self.project.uuid)

    def read_properties_from_dict(self, d):
        self.__source_proxy.item_specifier = Persistence.PersistentObjectSpecifier.read(d.get("source_uuid", None))
        self.original_expression = d.get("original_expression", self.original_expression)
        self.error_text = d.get("error_text", self.error_text)
        self.label = d.get("label", self.label)
        self.processing_id = d.get("processing_id", self.processing_id)

    def insert_model_item(self, container, name, before_index, item):
        if self.container:
            self.container.insert_model_item(container, name, before_index, item)
        else:
            container.insert_item(name, before_index, item)

    def remove_model_item(self, container, name, item, *, safe: bool=False) -> Changes.UndeleteLog:
        if self.container:
            return self.container.remove_model_item(container, name, item, safe=safe)
        else:
            container.remove_item(name, item)
            return Changes.UndeleteLog()

    def read_from_dict(self, properties):
        super().read_from_dict(properties)

    # override this so it can be short circuited if self.pending_project is available.
    # this is used to check inputs/outputs for circular dependencies before the computation is added to the project.
    def _get_related_item(self, item_specifier: Persistence.PersistentObjectSpecifier) -> typing.Optional[Persistence.PersistentObject]:
        related_item = super()._get_related_item(item_specifier)
        if related_item is None and self.pending_project:
            related_item = self.pending_project._get_related_item(item_specifier)
        return related_item

    @property
    def source(self):
        return self.__source_proxy.item

    @source.setter
    def source(self, source):
        self.__source_proxy.item = source
        self.source_specifier = source.project.create_specifier(source).write() if source else None

    def is_valid_with_removals(self, items: typing.Set) -> bool:
        for variable in self.variables:
            if variable.object_specifiers is not None:
                input_items = set(variable.input_items)
                # if removing items results in no bound items and at least some items are in bound items, computation
                # is no longer valid. all items of at least one set have been removed. do not remove computations with
                # empty bound items that already exist.
                if not (input_items - items) and input_items.intersection(items):
                    return False
        return True

    def __source_specifier_changed(self, name: str, d: typing.Dict) -> None:
        self.__source_proxy.item_specifier = Persistence.PersistentObjectSpecifier.read(d)

    @property
    def label(self) -> str:
        label = self._get_persistent_property_value("label")
        if not label:
            compute_class = _computation_types.get(self.processing_id)
            if compute_class:
                label = getattr(compute_class, "label", None)
        return label

    @label.setter
    def label(self, value: str) -> None:
        self._set_persistent_property_value("label", value)

    @property
    def error_text(self) -> typing.Optional[str]:
        return self._get_persistent_property_value("error_text")

    @error_text.setter
    def error_text(self, value):
        modified_state = self.modified_state
        self._set_persistent_property_value("error_text", value)
        self.modified_state = modified_state

    def __error_changed(self, name, value):
        self.notify_property_changed(name)
        self.computation_mutated_event.fire()

    def __label_changed(self, name, value):
        self.notify_property_changed(name)
        self.computation_mutated_event.fire()

    def add_variable(self, variable: ComputationVariable) -> None:
        self.insert_variable(len(self.variables), variable)

    def insert_variable(self, index: int, variable: ComputationVariable) -> None:
        self.insert_item("variables", index, variable)
        if self.persistent_object_context:
            self.__bind_variable(variable)
        self.variable_inserted_event.fire(index, variable)
        self.computation_mutated_event.fire()
        self.needs_update = True

    def remove_variable(self, variable: ComputationVariable) -> None:
        self.__unbind_variable(variable)
        index = self.item_index("variables", variable)
        self.remove_item("variables", variable)
        self.variable_removed_event.fire(index, variable)
        self.computation_mutated_event.fire()
        self.needs_update = True

    def create_variable(self, name: str=None, value_type: str=None, value=None, value_default=None, value_min=None, value_max=None, control_type: str=None, specifier: dict=None, label: str=None) -> ComputationVariable:
        variable = ComputationVariable(name, value_type=value_type, value=value, value_default=value_default, value_min=value_min, value_max=value_max, control_type=control_type, specifier=specifier, label=label)
        self.add_variable(variable)
        return variable

    def create_input_item(self, name: str, input_item: ComputationItem, *, property_name: str=None, label: str=None, _item_specifier: typing.Dict = None) -> ComputationVariable:
        # Note: _item_specifier is only for testing
        if input_item.items is not None:
            variable = ComputationVariable(name, items=input_item.items, label=label)
            self.add_variable(variable)
            return variable
        else:
            specifier = _item_specifier or DataStructure.get_object_specifier(input_item.item, input_item.type, allow_partial=False)
            secondary_specifier = DataStructure.get_object_specifier(input_item.secondary_item, allow_partial=False) if input_item.secondary_item else None
            variable = ComputationVariable(name, specifier=specifier, secondary_specifier=secondary_specifier, property_name=property_name, label=label)
            self.add_variable(variable)
            return variable

    def create_output_item(self, name: str, output_item: ComputationItem = None, *, label: str = None, _item_specifier: typing.Dict = None) -> ComputationOutput:
        # Note: _item_specifier is only for testing
        if output_item and output_item.items is not None:
            specifiers = [DataStructure.get_object_specifier(item.item, allow_partial=False) for item in output_item.items]
            result = ComputationOutput(name, specifiers=specifiers, label=label)
            self.append_item("results", result)
            if self.persistent_object_context:
                self.__bind_result(result)
            self.computation_mutated_event.fire()
            return result
        elif output_item:
            assert not output_item.type
            assert not output_item.secondary_item
            specifier = _item_specifier or DataStructure.get_object_specifier(output_item.item, allow_partial=False)
            result = ComputationOutput(name, specifier=specifier, label=label)
            self.append_item("results", result)
            if self.persistent_object_context:
                self.__bind_result(result)
            self.computation_mutated_event.fire()
            return result

    def remove_item_from_objects(self, name: str, index: int) -> None:
        variable = self._get_variable(name)
        variable.bound_items_model.remove_item(index)

    def insert_item_into_objects(self, name: str, index: int, input_item: ComputationItem) -> None:
        variable = self._get_variable(name)
        specifier = DataStructure.get_object_specifier(input_item.item, input_item.type, allow_partial=False)
        variable.bound_items_model.insert_item(index, self.__resolve_object_specifier(specifier))

    def list_item_removed(self, object) -> typing.Optional[typing.Tuple[int, int, dict]]:
        # when an item is removed from the library, this method is called for each computation.
        # if the item being removed matches a variable item, mark the computation as needing an update.
        # if the item is contained in a list variable, create the undelete entries and return them.
        # undelete_entries = list()
        for variable in self.variables:
            # check if the bound item matches the object. mark for update if so.
            # TODO: really needed?
            if variable.bound_item and variable.bound_item.value == object:
                self.needs_update = True
            # check if the bound item is a list and item in list matches the object. if so, create
            # an undelete entry for the variable. the undelete entry describes how to reconstitute
            # the list item.
            if variable.object_specifiers is not None:
                for index, (object_specifier, bound_item) in enumerate(zip(variable.object_specifiers, variable.bound_items_model.items)):
                    base_objects = bound_item.base_objects if bound_item else list()
                    if object in base_objects:
                        variables_index = self.variables.index(variable)
                        variable.bound_items_model.remove_item(index)
                        return (index, variables_index, copy.deepcopy(object_specifier))
        return None

    def __resolve_variable(self, object_specifier: dict) -> typing.Optional[ComputationVariable]:
        if object_specifier:
            uuid_str = object_specifier.get("uuid")
            uuid_ = Converter.UuidToStringConverter().convert_back(uuid_str) if uuid_str else None
            if uuid_:
                for variable in self.variables:
                    if variable.uuid == uuid_:
                        return variable
        return None

    data_source_types = ("data_source", "data_item", "xdata", "display_xdata", "cropped_xdata", "cropped_display_xdata", "filter_xdata", "filtered_xdata")

    def __resolve_object_specifier(self, specifier, secondary_specifier=None, property_name=None) -> typing.Optional[BoundItemBase]:
        """Resolve the object specifier.

        First lookup the object specifier in the enclosing computation. If it's not found,
        then lookup in the computation's context. Otherwise it should be a value type variable.
        In that case, return the bound variable.
        """
        bound_item = None
        variable = self.__resolve_variable(specifier)
        if not variable:
            if specifier and specifier.get("version") == 1:
                specifier_type = specifier["type"]
                project = self.container or self.pending_project
                if specifier_type == "data_source":
                    bound_item = BoundDataSource(project, specifier, secondary_specifier)
                elif specifier_type == "data_item":
                    bound_item = BoundDataItem(project, specifier)
                elif specifier_type == "xdata":
                    bound_item = BoundData(project, specifier)
                elif specifier_type == "display_xdata":
                    bound_item = BoundDisplayData(project, specifier, secondary_specifier)
                elif specifier_type == "cropped_xdata":
                    bound_item = BoundCroppedData(project, specifier, secondary_specifier)
                elif specifier_type == "cropped_display_xdata":
                    bound_item = BoundCroppedDisplayData(project, specifier, secondary_specifier)
                elif specifier_type == "filter_xdata":
                    bound_item = BoundFilterData(project, specifier, secondary_specifier)
                elif specifier_type == "filtered_xdata":
                    bound_item = BoundFilteredData(project, specifier, secondary_specifier)
                elif specifier_type == "structure":
                    bound_item = BoundDataStructure(project, specifier, property_name)
                elif specifier_type == "graphic":
                    bound_item = BoundGraphic(project, specifier, property_name)
        elif variable.specifier is None:
            bound_item = variable.bound_variable
        if bound_item and not bound_item.valid:
            bound_item.close()
            bound_item = None
        return bound_item

    @property
    def expression(self) -> str:
        return self.original_expression

    @expression.setter
    def expression(self, value: str) -> None:
        if value != self.original_expression:
            self.original_expression = value
            self.processing_id = None
            self.needs_update = True
            self.computation_mutated_event.fire()

    @classmethod
    def parse_names(cls, expression):
        """Return the list of identifiers used in the expression."""
        names = set()
        try:
            ast_node = ast.parse(expression, "ast")

            class Visitor(ast.NodeVisitor):
                def visit_Name(self, node):
                    names.add(node.id)

            Visitor().visit(ast_node)
        except Exception:
            pass
        return names

    def __resolve_inputs(self, api) -> typing.Tuple[typing.Dict, bool]:
        kwargs = dict()
        is_resolved = True
        for variable in self.variables:
            bound_object = variable.bound_item
            if bound_object is not None:
                resolved_object = bound_object.value if bound_object else None
                # in the ideal world, we could clone the object/data and computations would not be
                # able to modify the input objects; reality, though, dictates that performance is
                # more important than this protection. so use the resolved object directly.
                api_object = api._new_api_object(resolved_object) if resolved_object else None
                kwargs[variable.name] = api_object if api_object else resolved_object  # use api only if resolved_object is an api style object
                is_resolved = resolved_object is not None
            else:
                is_resolved = False
        for result in self.results:
            if result.specifier and not result.bound_item:
                is_resolved = False
            if result.specifiers and not all(result.bound_item):
                is_resolved = False
        return kwargs, is_resolved

    def evaluate(self, api) -> typing.Tuple[typing.Callable, str]:
        compute_obj = None
        error_text = None
        needs_update = self.needs_update
        self.needs_update = False
        if needs_update:
            kwargs, is_resolved = self.__resolve_inputs(api)
            if is_resolved:
                compute_class = _computation_types.get(self.processing_id)
                if compute_class:
                    try:
                        api_computation = api._new_api_object(self)
                        api_computation.api = api
                        compute_obj = compute_class(api_computation)
                        compute_obj.execute(**kwargs)
                    except Exception as e:
                        # import sys, traceback
                        # traceback.print_exc()
                        # traceback.format_exception(*sys.exc_info())
                        compute_obj = None
                        error_text = str(e) or "Unable to evaluate script."  # a stack trace would be too much information right now
                else:
                    compute_obj = None
                    error_text = "Missing computation (" + self.processing_id + ")."
            else:
                error_text = "Missing parameters."
            self._evaluation_count_for_test += 1
            self.last_evaluate_data_time = time.perf_counter()
        return compute_obj, error_text

    def evaluate_with_target(self, api, target) -> str:
        assert target is not None
        error_text = None
        needs_update = self.needs_update
        self.needs_update = False
        if needs_update:
            variables = dict()
            for variable in self.variables:
                bound_object = variable.bound_item
                if bound_object is not None:
                    resolved_object = bound_object.value if bound_object else None
                    # in the ideal world, we could clone the object/data and computations would not be
                    # able to modify the input objects; reality, though, dictates that performance is
                    # more important than this protection. so use the resolved object directly.
                    api_object = api._new_api_object(resolved_object) if resolved_object else None
                    variables[variable.name] = api_object if api_object else resolved_object  # use api only if resolved_object is an api style object

            expression = self.original_expression
            if expression:
                error_text = self.__execute_code(api, expression, target, variables)

            self._evaluation_count_for_test += 1
            self.last_evaluate_data_time = time.perf_counter()
        return error_text

    def __execute_code(self, api, expression, target, variables) -> typing.Optional[str]:
        code_lines = []
        g = variables
        g["api"] = api
        g["target"] = target
        l = dict()
        expression_lines = expression.split("\n")
        code_lines.extend(expression_lines)
        code = "\n".join(code_lines)
        try:
            compiled = compile(code, "expr", "exec")
            exec(compiled, g, l)
        except Exception as e:
            # print(code)
            # import sys, traceback
            # traceback.print_exc()
            # traceback.format_exception(*sys.exc_info())
            return str(e) or "Unable to evaluate script."  # a stack trace would be too much information right now
        return None

    def mark_update(self):
        self.needs_update = True
        self.computation_mutated_event.fire()

    @property
    def is_resolved(self):
        if not all(not v.specifier or v.bound_item for v in self.variables):
            return False
        for result in self.results:
            if result.specifier and not result.bound_item:
                return False
            if result.specifiers and not all(result.bound_item):
                return False
        return True

    def undelete_variable_item(self, name: str, index: int, specifier: typing.Dict) -> None:
        for variable in self.variables:
            if variable.name == name:
                variable.bound_items_model.insert_item(index, self.__resolve_object_specifier(specifier))

    def __bind_variable(self, variable: ComputationVariable) -> None:
        # bind the variable. the variable has a reference to another object in the library.
        # this method finds that object and stores it into the variable. it also sets up
        # listeners to notify this computation that the variable or the object referenced
        # by the variable has changed in a way that the computation needs re-execution,
        # and that the variable needs rebinding, which must be done from this class.

        def needs_update():
            self.needs_update = True
            self.computation_mutated_event.fire()

        self.__variable_changed_event_listeners[variable.uuid] = variable.changed_event.listen(needs_update)

        def rebind():
            self.needs_update = True
            self.__unbind_variable(variable)
            self.__bind_variable(variable)

        self.__variable_needs_rebind_event_listeners[variable.uuid] = variable.needs_rebind_event.listen(rebind)

        if variable.object_specifiers is not None:
            bound_items = list()
            for object_specifier in variable.object_specifiers:
                bound_item = self.__resolve_object_specifier(object_specifier)
                bound_items.append(bound_item)
            variable.connect_items(bound_items)
            variable.bound_item = BoundList(bound_items)
        else:
            variable.bound_item = self.__resolve_object_specifier(variable.variable_specifier, variable.secondary_specifier, variable.property_name)

    def __unbind_variable(self, variable: ComputationVariable) -> None:
        self.__variable_changed_event_listeners[variable.uuid].close()
        del self.__variable_changed_event_listeners[variable.uuid]
        self.__variable_needs_rebind_event_listeners[variable.uuid].close()
        del self.__variable_needs_rebind_event_listeners[variable.uuid]
        variable.bound_item = None
        variable.disconnect_items()

    def __bind_result(self, result: ComputationOutput) -> None:
        # bind the result. the result has an optional reference to another object in the library.
        # this method finds that object and stores it into the result. it also sets up
        # a listener to notify this computation that the result or the object referenced
        # by the result needs rebinding, which must be done from this class.

        def rebind():
            self.__unbind_result(result)
            self.__bind_result(result)
            self.computation_output_changed_event.fire()

        self.__result_needs_rebind_event_listeners[result.uuid] = result.needs_rebind_event.listen(rebind)

        result.bind(self.__resolve_object_specifier)

    def __unbind_result(self, result: ComputationOutput) -> None:
        self.__result_needs_rebind_event_listeners[result.uuid].close()
        del self.__result_needs_rebind_event_listeners[result.uuid]
        result.bound_item = None

    def bind(self, context) -> None:
        """Bind a context to this computation.

        The context allows the computation to convert object specifiers to actual objects.
        """

        # make a computation context based on the enclosing context.
        assert self.persistent_object_context

        # re-bind is not valid. be careful to set the computation after the data item is already in document.
        for variable in self.variables:
            assert variable.bound_item is None
        for result in self.results:
            assert result.bound_item is None

        # bind the variables
        for variable in self.variables:
            self.__bind_variable(variable)

        # bind the results
        for result in self.results:
            self.__bind_result(result)

        self.__is_bound = True

    def unbind(self):
        """Unlisten and close each bound item."""
        assert self.__is_bound

        for variable in self.variables:
            self.__unbind_variable(variable)
        for result in self.results:
            self.__unbind_result(result)

    def get_preliminary_input_items(self) -> typing.Set:
        input_items = set()
        for variable in self.variables:
            if variable.object_specifiers is not None:
                for object_specifier in variable.object_specifiers:
                    bound_item = self.__resolve_object_specifier(object_specifier)
                    if bound_item:
                        with contextlib.closing(bound_item):
                            input_items.update(bound_item.base_objects)
            else:
                bound_item = self.__resolve_object_specifier(variable.variable_specifier, variable.secondary_specifier, variable.property_name)
                if bound_item:
                    with contextlib.closing(bound_item):
                        input_items.update(bound_item.base_objects)
        return input_items

    def get_preliminary_output_items(self) -> typing.Set:
        output_items = set()
        for result in self.results:
            if result.specifier:
                bound_item = self.__resolve_object_specifier(result.specifier)
                if bound_item:
                    with contextlib.closing(bound_item):
                        output_items.update(bound_item.base_objects)
            elif result.specifiers is not None:
                for specifier in result.specifiers:
                    bound_item = self.__resolve_object_specifier(specifier)
                    if bound_item:
                        with contextlib.closing(bound_item):
                            output_items.update(bound_item.base_objects)
        return output_items

    @property
    def input_items(self) -> typing.Set:
        input_items = set()
        for variable in self.variables:
            input_items.update(variable.input_items)
        return input_items

    @property
    def direct_input_items(self) -> typing.Set:
        input_items = set()
        for variable in self.variables:
            input_items.update(variable.direct_input_items)
        return input_items

    @property
    def output_items(self) -> typing.Set:
        output_items = set()
        for result in self.results:
            output_items.update(result.output_items)
        return output_items

    def set_input_item(self, name: str, input_item: ComputationItem) -> None:
        for variable in self.variables:
            if variable.name == name:
                assert input_item.item
                assert input_item.type is None
                assert input_item.secondary_item is None
                assert input_item.items is None
                variable.specifier = DataStructure.get_object_specifier(input_item.item, allow_partial=False)

    def set_output_item(self, name:str, output_item: ComputationItem) -> None:
        for result in self.results:
            if result.name == name:
                if output_item and output_item.items is not None:
                    result.specifiers = [DataStructure.get_object_specifier(o.item, allow_partial=False) for o in output_item.items]
                else:
                    if output_item:
                        assert output_item.item
                        assert output_item.type is None
                        assert output_item.secondary_item is None
                    result.specifier = DataStructure.get_object_specifier(output_item.item, allow_partial=False) if output_item else None

    def get_input(self, name: str):
        for variable in self.variables:
            if variable.name == name:
                return variable.bound_item.value if variable.bound_item else None
        return None

    def get_output(self, name: str):
        for result in self.results:
            if result.name == name:
                if isinstance(result.bound_item, list):
                    return [bound_item.value for bound_item in result.bound_item]
                if result.bound_item:
                    return result.bound_item.value
        return None

    def set_input_value(self, name: str, value) -> None:
        for variable in self.variables:
            if variable.name == name:
                variable.value = value

    def get_input_value(self, name: str):
        for variable in self.variables:
            if variable.name == name:
                return variable.bound_variable.value if variable.bound_variable else None
        return None

    def get_input_base_items(self, name: str) -> typing.Optional[typing.Set]:
        for variable in self.variables:
            if variable.name == name:
                return variable.bound_item.base_objects if variable.bound_item else None
        return None

    def _get_variable(self, variable_name) -> typing.Optional[ComputationVariable]:
        for variable in self.variables:
            if variable.name == variable_name:
                return variable
        return None

    def _set_variable_value(self, variable_name, value):
        for variable in self.variables:
            if variable.name == variable_name:
                variable.value = value

    def _has_variable(self, variable_name: str) -> bool:
        for variable in self.variables:
            if variable.name == variable_name:
                return True
        return False

    def _clear_referenced_object(self, name: str) -> None:
        for result in self.results:
            if result.name == name:
                if result.bound_item:
                    self.__unbind_result(result)
                index = self.item_index("results", result)
                self.remove_item("results", result)
                # self.result_removed_event.fire(index, result)
                self.computation_mutated_event.fire()
                self.needs_update = True

    def _get_reference(self, name: str):
        for result in self.results:
            if result.name == name:
                return result
        return None

    def update_script(self, processing_descriptions) -> None:
        processing_id = self.processing_id
        processing_description = processing_descriptions.get(processing_id)
        if processing_description:
            expression = processing_description.get("expression")
            if expression:
                src_names = list()
                source_dicts = processing_description["sources"]
                for i, source_dict in enumerate(source_dicts):
                    src_names.append(source_dict["name"])
                script = xdata_expression(expression)
                script = script.format(**dict(zip(src_names, src_names)))
                self._get_persistent_property("original_expression").value = script

# for computations

_computation_types = dict()

def register_computation_type(computation_type_id: str, compute_class: typing.Callable) -> None:
    _computation_types[computation_type_id] = compute_class

# for testing

def xdata_expression(expression: str=None) -> str:
    return "import numpy\nimport uuid\nfrom nion.data import xdata_1_0 as xd\ntarget.xdata = " + expression

def data_expression(expression: str=None) -> str:
    return "import numpy\nimport uuid\nfrom nion.data import xdata_1_0 as xd\ntarget.data = " + expression
